/** File Cache.java
* Classe astratta Java che implementa 
* una cache vista da un client.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;
import java.util.*;

public abstract class Cache implements ICache, ICacheDataLoader
{
   // membri
   private ICacheChain chain = null;
   private CacheConstants cc = null;
   private Socket sock = null;
   private DataInputStream datain = null;
   private DataOutputStream dataout = null;
   private Thread snooper = null;
   private String cachename = null;
   private String servername = null;
   private long cacheID = 0;
   private InetAddress serveraddr = null;
   private int serverport = 0;
   private Boolean server_ok = null;
   private Boolean lock = null; // lock di server_ok
   private CacheCLMDNR clogic = null;
   
   // il thread di snoop
   private class Snooper implements Runnable
   {
      private long serverID = 0; // servir ad identificare il server
      private ServerSocket listener;
      private DataInputStream datain = null;
      private DataOutputStream dataout = null;
      private Socket snoopsock = null;
      private Event connected;
      private boolean connectiondone = false;
      
      public Snooper(ServerSocket listener) throws IOException
      {
         connected = new Event();
         connected.Reset();
         this.listener = listener;
      }
      
      public synchronized boolean isConnected()
      { return connectiondone; }

      public void waitForConnection()
      {
         connected.waitEvent();   
      }
      
      public void run()
      {
         try{
         // buffer per il nome del server 
         byte[] buffer = new byte[cc.SERVER_NAME_SIZE];
         boolean cont = true;
         
         // all'inizio il thread snooper deve attendere
         // che il server si identifichi
         // pertanto il tempo di attesa pu risultare elevato
         listener.setSoTimeout(cc.WAIT_FOR_SERVER_RESPONSE); // attesa MOOLTO grande
         while(cont)
         {
            snoopsock = listener.accept();
            // settaggio opzioni socket
            //debug opzione eliminata
            //snoopsock.setTcpNoDelay(cc.TCPNODELAY);
            snoopsock.setSoTimeout(cc.TIMEOUT);
            snoopsock.setSoLinger(cc.LINGER, cc.LINGERTIME);
            // imposto gli stream
            datain = new DataInputStream( snoopsock.getInputStream() );
            // rileggi l'ID della cache
            long retainedID = datain.readLong();
            // se  giusto questo  l'ultimo ciclo
            cont = ( retainedID != cacheID );
            // ed allora si deve anche leggere il nome del server
            if(!cont) 
            {   
               servername = datain.readUTF();
               //debug
               System.err.println("Server name is "+servername);
               // e si deve leggere anche l'id del server
               serverID = datain.readLong(); 
               //debug
               System.err.println("Server ID is "+serverID);
            }
            snoopsock.close();
            snoopsock = null;
         }
         
         // se siamo qui allora il server si  identificato
         // percui si pu iniziare una iterazione.
         // non c' bisogno di sincronizzarsi in quanto 
         // siamo ancora in fase di costruzione della cache
         // e fino alla waitForConnection() la cache NON
         // usa la variabile server_ok
         server_ok = new Boolean(true);
         connectiondone = true;
         connected.Happen();
         
         // ed ora possiamo ciclare per risolvere gli snoop
         // ogni tanto per dovremo anche verificare che 
         // lo snooper non venga fermato
         listener.setSoTimeout(cc.ACCEPT_POLLING_TIME);
         while(!StopSnooper())
         {
            boolean checkoutsrv = false;
            // server in parte parallelo
            // IPOTESI : NON CI SONO MAI TANTI
            // CICLI DI SNOOP PENDENTI
            try
            {
               snoopsock = null;
               snoopsock = listener.accept();
            }
            catch( InterruptedIOException ie ) // ogni 3 secondi
            { checkoutsrv = true; }
            // se si deve controllare il server allora snoopsock
            // non punta a nulla a causa dell' eccezione
            if(checkoutsrv)
               if(ServerOK()) // se il server  a posto rimetto a false chkoutsrv
               {
                  checkoutsrv = false;
                  if(null== snoopsock) continue;
               }
               else // il server ha finito dunque si deve uscire
               {
                  listener.close();
                  return;
               }
            else // non  avvenuta alcuna eccezione: snoopsock  valida
               if(!ServerOK()) // se si deve terminare liquida il server
               {
                 listener.close();
                 datain = new DataInputStream( snoopsock.getInputStream() );
                 dataout = new DataOutputStream( snoopsock.getOutputStream() );
                 datain.readLong();// legge l'id
                 datain.readShort();// cassa il comando
                 dataout.writeShort(cc.MISS);// cassazione
                 dataout.flush();
                 datain.readShort(); // cassa l'ACK
                 snoopsock.close();
                 return;
              }
            
            // se siamo qui il server  a posto e snoopsock pure
            // settaggio opzioni socket
            snoopsock.setTcpNoDelay(cc.TCPNODELAY);
            snoopsock.setSoTimeout(cc.TIMEOUT);
            snoopsock.setSoLinger(cc.LINGER, cc.LINGERTIME);
            // imposto gli stream
            datain = new DataInputStream( snoopsock.getInputStream() );
            dataout = new DataOutputStream( snoopsock.getOutputStream() );

            // leggiamo l'ID di chi richiede lo snoop
            long ID = datain.readLong();
            // se non  il server, cassiamo la richiesta
            if(ID != serverID) 
            {
               snoopsock.setSoLinger(true, 0); // hard close
               snoopsock.close();
               // torniamo ad inizio ciclo
               continue;
            }
            
            // se l'ID  giusta, leggiamo la richiesta
            short command = datain.readShort();
            if(command == cc.SERVER_IS_CLOSING)
            {
               SnoopOperation operation =
                         new SnoopOperation( snoopsock, 
                                             datain,
                                             dataout,
                                             cc.SERVER_NEEDS_FLUSHING );
               operation.run();// non c' concorrenza per il flush
               StopInteractionWithServer();// ferma ogni ulteriore comando
                                           // lanciato dal client
             
            }
            else
            {
               Thread operation = new Thread(
                              new SnoopOperation( snoopsock, 
                                                  datain,
                                                  dataout,
                                                  command ) );
               
               operation.start();// i normali snoop sono concorrenti
            }
         } // siamo fuori : il server non far pi nulla
         listener.close();
         } // try
         catch(IOException e)
         {
            System.err.println("Error during snooping allowance. Shutting down...");
            System.err.println(e);
            e.printStackTrace();
            StopInteractionWithServer();
         }
      }  
      
   } // Snooper private class
   
   private class SnoopOperation implements Runnable
   {
      private Socket socket;
      private DataInputStream datain;
      private DataOutputStream dataout;
      private short command;
      private ICacheElem WBelem;// indica, se != null che siamo nella fase di WB
      private ICacheElem[] WBelems;// indica, se != null che siamo nella fase di WB di un flush
      private int flshindex; // indica gli elementi non ancora flushati
      public SnoopOperation( Socket socket, 
                             DataInputStream datain,
                             DataOutputStream dataout,
                             short command )
      {
         this.socket = socket;
         this.datain = datain;
         this.dataout = dataout;
         this.command = command;
         WBelem = null ; // siamo nella fase iniziale della operazione
         WBelems = null; // idem
         flshindex = 0;
      }
      
      public void run()
      {
         // elenco dei gestori degli eventi
         try
         {
            switch(command)
            {
               case cc.SNOOP: snoop(); break;
               case cc.SNOOP_INV: snoop_inv(); break;
               case cc.SERVER_NEEDS_FLUSHING: flushings(); break;
            }
         } catch(IOException e)
         {
            // Some buddy has wreck out of sightness
            System.err.println(e);
            System.err.println("The server can't receive the datas it has requested; Disconnecting...");
            try
            {
               if(WBelem != null) MissedModifiedData(WBelem);
               else if(WBelems != null) MissedModifiedDataFlush(WBelems, flshindex);
                    else ServerShoutedDown();
            }
            catch(ServerDownException exc ){}
         }
         try{socket.close();} catch(IOException e){}
      }
      
      private void snoop() throws IOException
      {
         ITag fingerhang = readTag(datain);
         ICacheElem casualty;
         ICacheElem fatality;
         synchronized(chain) // azione atomica sulla catena
         {
            // considera se il dato  in cache
            casualty = chain.getData(fingerhang);
            // e se c'  in stato M?... e in tal caso condividilo
            fatality = (casualty!=null)? chain.shareData(fingerhang): null;
            //debug
            System.err.println(chain.toString());
         }
         if( casualty == null ) 
         {
            dataout.writeShort(cc.MISS);// il dato non  in cache
            dataout.flush();
         }
         else if( fatality == null ) 
              {
                 dataout.writeShort(cc.HIT);// il dato  in cache ma non  modificato
                 dataout.flush();
              }
              else
              { 
                 dataout.writeShort(cc.HITM);
                 WBelem = fatality; // fase di writeback iniziata
                 fatality.writeImage(dataout);
                 dataout.flush();
              }// GOTCHA!!!   
         // wait for the filthy acknowledge
         datain.readShort();
      }
      
      private void snoop_inv() throws IOException
      {
         ITag fingerhang = readTag(datain);
         ICacheElem casualty;
         ICacheElem fatality;
         synchronized(chain) // azione atomica sulla catena
         {
            // considera se il dato  in cache
            casualty = chain.getData(fingerhang);
            // e se c'  in stato M?... e in tal caso HAVE A NICE FEAST!!!
            fatality = (casualty!=null)? chain.invalidData(fingerhang): null;
            //debug
            System.err.println(chain.toString());
         }
         if( casualty == null ) 
         {
            dataout.writeShort(cc.MISS);// il dato non  in cache
            dataout.flush();
         }
         else if( fatality == null ) 
              {
                 dataout.writeShort(cc.HIT);// il dato  in cache ma non  modificato
                 dataout.flush();
              }
              else
              { 
                 dataout.writeShort(cc.HITM);
                 WBelem = fatality; // fase di writeback iniziata
                 fatality.writeImage(dataout);
                 dataout.flush();
              }// GOTCHA!!!   
         // wait for the filthy acknowledge
         datain.readShort();
      }
      
      private void flushings() throws IOException
      {
         ICacheElem[] victims = chain.flushData();
         //debug
         System.err.println(chain.toString());
         int bodycount = ( victims == null )? 0 : victims.length ;
         if( bodycount == 0 ) 
         {
            dataout.writeShort(cc.MISS);
            dataout.flush();
         }
         else
         {
            // inviamo tutti i dati non inviati
            dataout.writeShort(cc.HITM);
            WBelems = victims;
            dataout.writeInt(bodycount);
            for( int j = 0; j < bodycount ; j++)
            {
               victims[j].writeImage(dataout);
               flshindex++;
            }
            dataout.flush();
         }
         // wait for the filthy acknowledge
         datain.readShort();
      }
   
   } // SnoopOperation Thread
   
            
   // costruttore
   public Cache(String name,
                long cacheID,
                int truesize,
                String server_logic_addr,
                int serverport) throws IOException
   {
      // creo la logica di interazione e consistenza
      clogic = createConsistencyLogic();
      // inizialmente non si ha interazione col server
      server_ok = new Boolean(false);
      lock = new Boolean(true);
      
      this.cachename = name;
      this.cacheID = cacheID;
      this.serverport = serverport;
      cc = new CacheConstants();
      
      // il size della cache  il pi grande tra il default e quello del file config
      truesize = (truesize > cc.DEFAULT_CACHE_SIZE) ?
                      truesize: cc.DEFAULT_CACHE_SIZE;
      System.out.println("Setting cache size to "+truesize);
      // creo la catena degli elementi
      chain = new CacheChain(truesize);
      System.out.println("Searching for server "+server_logic_addr+":"+serverport);
      // ottengo l'IP del server
      serveraddr = InetAddress.getByName(server_logic_addr);
      // creo la socket passiva dove il server lancer gli snoop
      ServerSocket snoopsock = new ServerSocket(0);
      int snoopport = snoopsock.getLocalPort();
      // creo il thread di snoop
      // NOTA: DA QUI A DOPO waitForConnection() 
      // la variabile server_ok NON DEVE ESSERE USATA
      Snooper snp = new Snooper( snoopsock );
      snooper = new Thread( snp );
      snooper.start();
      // invio la richiesta di connessione
      System.out.println("Contacting server "+serveraddr+":"+serverport);
      sock = new Socket(serveraddr, serverport);
      // imposta le opzioni della socket
      sock.setTcpNoDelay(cc.TCPNODELAY);
      sock.setSoTimeout(cc.TIMEOUT);
      sock.setSoLinger(cc.LINGER, cc.LINGERTIME);
      // preleva le info sugli streams
      datain = new DataInputStream( sock.getInputStream() );
      dataout = new DataOutputStream( sock.getOutputStream() );
      // inizio di richiesta di connessione
      System.out.println("Sending local settings...");
      dataout.writeShort(cc.CLIENT_BEGINS_SESSION); 
      // invio al server del proprio numero di porta ed il proprio ID
      System.out.println("Sending to server local port address "+snoopport);
      dataout.writeInt(snoopport);
      dataout.writeLong(cacheID);
      dataout.flush();
      // attendi che il server abbia effettuato la connessione
      System.out.println("Awaiting server reply at port "+snoopport);
      snp.waitForConnection();
      // se siamo qui, tutto  andato a posto
      // e la variabile server_ok pu ora essere usata
   }
      
   // metodi locali usati dallo snooper
   private boolean ServerOK()
   {
      synchronized(lock)// accesso escusivo a server_ok
      {
         return server_ok.booleanValue();
      }
   }
   
   private boolean StopSnooper()
   { return !ServerOK(); }
   
   private synchronized void StopInteractionWithServer()
   {
      try{sock.close(); }
      catch(IOException e){}
      
      synchronized(lock)// accesso escusivo a server_ok
      {
         server_ok = new Boolean(false);
      }
   }
   
   private void ServerShoutedDown() throws ServerDownException
   {
      StopInteractionWithServer();
      throw new ServerDownException(servername);
   }
   
   private void MissedModifiedData( ICacheElem data ) throws ServerDownException
   {
      StopInteractionWithServer();
      System.err.println("Followings are missing data...");
      System.err.println(data);
      throw new ServerDownException( servername );
   }
   
   private void MissedModifiedDataFlush( ICacheElem[] data, int notFlushed ) throws ServerDownException
   {
      StopInteractionWithServer();
      System.err.println("Followings are missing data...");
      for(int i = notFlushed; i<data.length ; i++)
         System.err.println(data[i]);
      throw new ServerDownException( servername );
   }
   
   private void SendAckNException( CacheException e ) throws CacheException,
                                                             IOException
   {
      dataout.writeShort(cc.ACK);
      dataout.flush();
      throw e;
   }
   
   // metodi di ICache
   public ICacheElem readLocal( ITag tag ) throws 
                                           NotInCacheException
   {
      ICacheElem elem = chain.getData(tag);
      if ( elem == null ) throw new NotInCacheException(tag, this );
      return elem;
   }
   
   public synchronized ICacheElem read( ITag tag, boolean cacheit ) throws
                                      CacheException
                                      //LinefillMissException,
                                      //ServerDownException,
                                      //LinefillHaltedException,
                                      //ServerPermissionException
   {
      ICacheElem elem = chain.getData(tag);
      ICacheElem prev = null; // eventuale elemento modified da writebackkare
      if( elem == null ) // se il dato non  in cache
      {
         if(!ServerOK()) ServerShoutedDown();
         try
         {
            short command = cacheit? cc.LINEFILL_CACHEABLE : cc.LINEFILL_UNCACHEABLE;
            boolean shared = true; // indicher lo stato iniziale del dato
            dataout.writeShort(command);
            tag.writeImage(dataout);
            dataout.flush();
            short result = datain.readShort();
            switch(result)
            {
            // errori : invia ACK e lancia un eccezione
            case cc.LF_ANS_MISS : 
               SendAckNException( new LinefillMissException(tag, servername) );
               break;
            case cc.LF_ANS_HALTED :               
               SendAckNException( new LinefillHaltedException(tag, servername) );
               break;
            case cc.ACCESS_VIOLATION :
               SendAckNException( new ServerPermissionException(tag, servername) );
               break;
            // non cacheable : ritorna semplicemente l'elemento
            // ed invia l'ACK
            case cc.LF_ANS_UNCACHEABLE : 
               elem = readCacheElem(datain);
               dataout.writeShort(cc.ACK);
               dataout.flush();
               return elem;
            case cc.LF_ANS_CACHE_EXCL : shared = false; break;
            case cc.LF_ANS_CACHE_SHARE : shared = true ; break;
            }
            // se siamo qui allora dobbiamo finire di gestire uno degli ultimi due casi
            // leggendo il dato e inserendolo in cache
            elem = readCacheElem(datain);
            prev = chain.insertData( elem, shared );
            //debug
            System.err.println(chain.toString());
            // se c' stato un writeback dobbiamo scrivere il dato precedente
            // sul server
            if(prev!=null)
            {
               dataout.writeShort(cc.LINEFILL_CAUSED_WRITEBACK);
               prev.writeImage(dataout);
               dataout.flush();
               short ack = datain.readShort();
               // se ci sono dei problemi lancia un eccezione
               if(ack != cc.ACK) MissedModifiedData(prev); 
               prev = null;
            }
            // e se non c' WB indichiamo solo l'ACK
            else { dataout.writeShort(cc.ACK); dataout.flush(); }
            // in ogni caso infine ritorniamo il dato letto
            return elem;
         }
         catch( IOException e )
         {
            System.err.println("Server is not responing. Shutting down...");
            System.err.println(e);
            // se l'errore  avvenuto nella fasew di writeback
            // stampa il dato perso e lancia una eccezione
            if(prev!=null) MissedModifiedData(prev);
            else ServerShoutedDown(); // altrimenti solo eccezione
            return null; // uno stub mai eseguito
         }
      }   
      // altrimenti, se il dato era in cache...
      else return elem ;
   }
                                             
   public synchronized void write( ICacheElem elem, boolean forceWT ) throws 
                                        CacheException
                                        //NackWriteException,
                                        //ServerDownException,
                                        //ServerPermissionException
   {
      try
      {
         // vediamo se il nodo  in cache
         // pertanto sincronizziamoci sulla catena 
         // per evitare snoop
         synchronized(chain)
         {
            int status = chain.getStatus(elem.getTag());
            if((status!=cc.I)&& !forceWT) // caso di hit o non WT forzato
               switch(status)
               {
               case cc.S: // dobbiamo informare il server della scrittura
                  if(!ServerOK()) ServerShoutedDown();
                  // inviamo dunque una richiesta di passaggio ad uso 
                  // esclusivo
                  dataout.writeShort(cc.WRITE_EXCLUSIVE);
                  elem.getTag().writeImage(dataout); // serve a capire a quale dato ci riferiamo
                  dataout.flush();
                  // rileviamo la risposta
                  switch(datain.readShort())
                  {
                  case cc.NACK: // il server non  in grado di scrivere
                     throw new NackWriteException(elem.getTag(), servername); 
                     //return; // esci - VERY IMPORTANT
                     //break;
                  case cc.ACCESS_VIOLATION: // il server non vuole che noi scriviamo
                     SendAckNException( new ServerPermissionException(elem.getTag(), servername) );
                     //return; // esci - VERY IMPORTANT
                     //break;
                  case cc.ACK: // tutto a posto, possiamo scrivere localmente
                     break; // andiamo al caso successivo !!!
                  } 
                  // niente break !! scriviamo localmente
               case cc.E: // scriviamo solo localmente
               case cc.M: // scriviamo solo localmente
                  chain.invalidData(elem.getTag()); // libera uno slot
                  chain.insertData(elem, true); // scrive il dato
                  chain.setStatus(elem.getTag(), cc.M); // imposta lo stato
                  return; // esci - VERY IMPORTANT
               } // fine switch ed if : status == cc.I o forceWT vero
               
            // in caso di WT forzato se il dato era in chain
            // aggiornalo al nuovo valore come exclusive
            if(forceWT && (null!=chain.getData(elem.getTag())))
            {
               chain.invalidData(elem.getTag());
               chain.insertData(elem, false);
            }
         } // synchronized(chain) : liberiamo la catena
         
         // caso di miss o WT forzato - Writethrough
         if(!ServerOK()) ServerShoutedDown();
         // inviamo dunque una richiesta di WT
         dataout.writeShort(cc.WRITETHROUGH);
         elem.getTag().writeImage(dataout); // serve a capire a quale dato ci riferiamo
         dataout.flush();
         // rileviamo la risposta
         switch(datain.readShort())
         {
         case cc.NACK: // il server non  in grado di scrivere
            throw new NackWriteException(elem.getTag(), servername); 
            //return; // esci - VERY IMPORTANT
            //break;
         case cc.ACCESS_VIOLATION: // il server non vuole che noi scriviamo
            SendAckNException( new ServerPermissionException(elem.getTag(), servername) );
            return; // esci - VERY IMPORTANT
            //break;
         case cc.ACK: // tutto a posto, possiamo scrivere il dato
            elem.writeImage(dataout);
            dataout.flush();
            // retry la risposta : se c' stato un fallimento del 
            // passaggio a modalit esclusiva il server ritorna NACK
            if(cc.ACK!=datain.readShort()) 
               throw new NackWriteException(elem.getTag(), servername); 
            break; 
         } 
         
      } // try
      catch(IOException e)
      {
         System.err.println("Server is not responing. Shutting down...");
         System.err.println(e);
         // il server non  in grado di accettare la nostra
         // scrittura che dunque fallisce
         MissedModifiedData(elem);
      }
   }
   
   public synchronized void update( IUpdateElem news, boolean forceWT ) throws
                                        CacheException
                                        //InconsistentUpdateException,
                                        //NackWriteException,
                                        //ServerDownException,
                                        //ServerPermissionException
   {
      try
      {
         // cerchiamo il dato in cache : se c'  dobbiamo fare le verifiche                                        
         // in locale
         synchronized(chain)
         {
            int status = chain.getStatus( news.getTag() );
            if(status != cc.I) // se c'
            {
               // verifica la bont del cambiamento
               ICacheElem newElem = updateIsCorrect( chain.getData( news.getTag() ), news );
               // se siamo qui il cambiamento  possibile
               // allora se lo stato  cc.S interroghiamo il server
               // sul cambio. E' come se facessimo una write
               // perch in entrambi i casi in SHARED si ha solo
               // un write-exclusive.
               // NOTA: le cose vanno bene anche in stato E ed M !!!
               write(newElem, forceWT);
               return; // se siamo qui  tutto ok
            }
             
            // caso di miss
            // abbiamo un writethrough 
            // dunque rilasciamo la catena
         } // synchronized(chain)
         if(!ServerOK()) ServerShoutedDown();
         dataout.writeShort(cc.UPDATE);
         // abbiamo un passo in meno rispetto al caso 
         // precedente di write
         news.writeImage(dataout); // invio le info di update
         dataout.flush();
         switch(datain.readShort()) // legge il risultato
         {
         case cc.ACK: return; // tutto ok
         case cc.NACK: // non si riesce a scrivere
            throw new NackWriteException(news.getTag(), servername); 
            //return; // esci - VERY IMPORTANT
            //break;
         case cc.ACCESS_VIOLATION:
            SendAckNException( new ServerPermissionException(news.getTag(), servername) );
            return; // esci - VERY IMPORTANT
            //break;
         case cc.CANT_UPDATE_INCONSISTENCE:
            SendAckNException( new InconsistentUpdateException(news.getTag(), news) );
            return; // esci - VERY IMPORTANT
            //break;
         }
      } // try
      catch(IOException e)
      {
         System.err.println("Server is not responing. Shutting down...");
         System.err.println(e);
         // il server non  in grado di accettare la nostra
         // scrittura che dunque fallisce
         System.err.println("Following is a missing update...\n" + news);
         ServerShoutedDown();
      }
   
   } // update()       
                 
   public synchronized  void flush() throws 
                       CacheException
                       //NackWriteException,
                       //ServerDownException,
                       //ServerPermissionException
   {
      // preleva i dati modified
      ICacheElem[] dirties = chain.flushData();
      if( dirties == null ) return ;
      // se ci sono, li invia in writethrough
      if(!ServerOK()) ServerShoutedDown();
      ICacheElem elem;
      int index = 0;
      for( ; index < dirties.length ; index ++ )
      {   elem = dirties[index];
      try
      {
         // inviamo dunque una richiesta di WT
         dataout.writeShort(cc.WRITETHROUGH);
         elem.getTag().writeImage(dataout); // serve a capire a quale dato ci riferiamo
         dataout.flush();
         // rileviamo la risposta
         switch(datain.readShort())
         {
         case cc.NACK: // il server non  in grado di scrivere
            throw new NackWriteException(elem.getTag(), servername); 
            //return; // esci - VERY IMPORTANT
            //break;
         case cc.ACCESS_VIOLATION: // il server non vuole che noi scriviamo
            SendAckNException( new ServerPermissionException(elem.getTag(), servername) );
            return; // esci - VERY IMPORTANT
            //break;
         case cc.ACK: // tutto a posto, possiamo scrivere il dato
            elem.writeImage(dataout);
            dataout.flush();
            // retry la risposta
            if(cc.ACK!=datain.readShort()) 
               throw new NackWriteException(elem.getTag(), servername); 
            break; 
         } 
         
      } // try
      catch(IOException e)
      {
         System.err.println("Server is not responing. Shutting down...");
         System.err.println(e);
         // il server non  in grado di accettare la nostra
         // scrittura che dunque fallisce
         MissedModifiedDataFlush(dirties, index + 1);
      }
      } // for
   
   } // flush
                       
   public String toString()
   {
      return "Cache name : "+cachename+
             "\nCache ID : "+cacheID+
             "\nServer port : " + serverport+
             (ServerOK()?"\nThe server is running.\n":
                          "\nThe server is dead.\n") ;
   }
   
   // metodi di ICacheDataLoader
   public ITag readTag( DataInputStream datain ) throws IOException
   {
      return clogic.readTag(datain);
   }
   
   public IUpdateElem readUpdateElem( DataInputStream datain ) throws IOException
   {
      return clogic.readUpdateElem(datain);
   }
   
   public ICacheElem readCacheElem( DataInputStream datain ) throws IOException
   {
      return clogic.readCacheElem(datain);
   }
   
   public ICacheElem updateIsCorrect( ICacheElem oldelem, IUpdateElem news ) throws InconsistentUpdateException
   {
      return clogic.updateIsCorrect(oldelem,news);
   }
   
   // metodi astratti
   protected abstract CacheCLMDNR createConsistencyLogic();
   
   // funzione di chiusura
   public synchronized void closeSession() 
   {
      // se il server non  down
      if(ServerOK()) {
      try
      {
         // anzitutto flushiamo la cache
         flush();
         // poi indichiamo al server la nostra
         // volont di ritirarci
         dataout.writeShort(cc.CLIENT_DISCONNECTING);
         dataout.flush();
         // ritiriamo l'ACK
         datain.readShort();
         // blocchiamo ogni ulteriore snoop
         StopInteractionWithServer();
      }
      catch (Exception e)
      {
         System.err.println("Exception on closing:");
         System.err.println(e);
      }
      } // se il server era down, non far nulla
   
   } // closeSession()
   
   // distruttore virtuale
   protected void finalize() throws Throwable
   {
      closeSession();
   }
      
} // Cache

