/** File CacheConstants.java
* Classe Java che definisce 
* alcune costanti utili.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

public class CacheConstants
{
   // MESI STATES
   public static final int M = 3;
   public static final int E = 2;
   public static final int S = 1;
   public static final int I = 0;
   
   // dimensioni della cache a default
   public static final int DEFAULT_CACHE_SIZE = 3;
   
   // no. tentativi di snoop su server
   public static final int SNOOP_TRIES_NO = 3;
   
   // dimensioni del nome del server
   public static final int SERVER_NAME_SIZE = 56;
   // comandi di rete della cache 
   // risposta positiva
   public static final short ACK = 0;
   // risposta negativa
   public static final short NACK = 0x7FFF;
   // violazione di sicurezza di accesso
   public static final short ACCESS_VIOLATION = 0x3FFF;
   // richiesta di linefill da mettere in cache
   public static final short LINEFILL_CACHEABLE = 5;
   // richiesta di linefill da non mettere in cache
   public static final short LINEFILL_UNCACHEABLE = 10;
   // risposta al linefill : dato non trovato su nessun server
   public static final short LF_ANS_MISS = 11;
   // risposta al linefill : linefill interrotto da un altra operazione
   public static final short LF_ANS_HALTED = 9;
   // risposta al linefill : il dato non deve essere messo in cache
   public static final short LF_ANS_UNCACHEABLE = 12;
   // risposta al linefill : il dato deve essere messo in cache in stato E
   public static final short LF_ANS_CACHE_EXCL = 13;
   // risposta al linefill : il dato deve essere messo in cache in stato S
   public static final short LF_ANS_CACHE_SHARE = 14;
   // richiesta di writeback a seguito di linefill al server
   public static final short LINEFILL_CAUSED_WRITEBACK = 8;
   // richiesta di writethrough al server
   public static final short WRITETHROUGH = 15;
   // richiesta di uso esclusivo di un dato al server
   public static final short WRITE_EXCLUSIVE = 25;
   // inizio di un ciclo di snoop dal server al client
   public static final short SNOOP = 35;
   // risposta allo snoop : dato modified
   public static final short HITM = 45;
   // risposta allo snoop : dato non modified
   public static final short HIT = 55;
   // risposta allo snoop : dato non trovato
   public static final short MISS = 65;
   // inizio di un ciclo di snoop con invalidazione dal server al client
   public static final short SNOOP_INV = 75;
   // terminazione del client
   public static final short CLIENT_DISCONNECTING = 80;
   // terminazione del server
   public static final short SERVER_IS_CLOSING = 85;
   // inizio di un ciclo di snoop dal server al client ove il server chiede al client di flusharsi
   public static final short SERVER_NEEDS_FLUSHING = 95;
   // inizio di una sessione di interazione tra client e server
   public static final short CLIENT_BEGINS_SESSION = 105;
   // comando di update di undato sul server
   public static final short UPDATE = 1085;
     
   // comandi tra server
   public static final short SLAVE_IS_CALLING = 1000;
   public static final short SLAVE_IS_CLOSING = 1001;
   public static final short MASTER_IS_CLOSING = 1002;
   
   public static final short SERVER_BEGINS_SESSION = 1005;
   public static final short CAN_DO = 1095;
   public static final short CANT_DO_BECAUSE_COLLISION = 1105;
   public static final short CANT_UPDATE_INCONSISTENCE = 1106;
   public static final short DO_ACTION = 1110;
   public static final short UNDO_ACTION = 1111;
   public static final short SERVER_INV = 1115;
   public static final short GET_DATA = 1125;
   public static final short WRITE_DATA = 1126;
   public static final short DETECT_DATA = 1127;
   public static final short INV_DATA = 1128;
   public static final short DATA_MISS = 1135;
   public static final short DATA_MISS_SUGGEST = 1145;
   public static final short DATA_PRESENT = 1155;
   
   // opzioni sulle sockets
   public static final boolean TCPNODELAY = true;
   public static final boolean LINGER = true;
   public static final int TIMEOUT = 500000;// millisecs
   public static final int LINGERTIME = 500;// secs
   public static final int WAIT_FOR_SERVER_RESPONSE = 1000000; // ms
   public static final int ACCEPT_POLLING_TIME = 3000; // msecs
   public static final int MAX_WAIT_TIME_AFTER_COLLISION = 1000; // msecs
   public static final int SLEEPY_TIME = 1500; // ms
   
   // funzioni globali generiche
   public static byte[] BufferAdjust(byte[] startbuf, int finalsize)
   {
      byte[] finalbuf = new byte[finalsize];
      int size = ( finalsize < startbuf.length ) ? finalsize : startbuf.length ;
      for( int i = 0 ; i < size ; i++ )
         finalbuf[i] = startbuf[i];
      while( size < finalsize )
      {
         finalbuf[size] = (byte)0;
         size++;
      }
      return finalbuf;
   }
   
   public static byte[] StringToBuffer( String s, int buflen )
   {
      return BufferAdjust( s.getBytes(), buflen );
   }
   
} // CacheConstants
