/** File CacheSnooperThread.java
* Classe Java che implementa  
* un thread che dal server snoopa gli altri clients.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;

public class CacheSnooperThread extends Thread
{
   private int index = 0;
   private short[] res = null;
   private ICacheElem[] elemres = null;
   private ICacheElem newelem = null;
   private ITag tag = null;
   private CacheServer owner = null;
   private boolean inv_snoop = false;
   private CacheClientInfo target = null;
   private boolean open = false;
   private Socket sock = null;
   private DataInputStream datain = null;
   private DataOutputStream dataout = null;
   
   public CacheSnooperThread( int index,
                              short[] res, 
                              ICacheElem[] elemres,
                              CacheClientInfo target,
                              ICacheElem newelem,
                              CacheServer owner,
                              boolean inv_snoop )
   {
      this.index = index;
      this.res = res;
      this.elemres = elemres;
      this.target = target;
      this.newelem = newelem;
      tag = newelem.getTag();
      this.owner = owner;
      this.inv_snoop = inv_snoop;
   }
   
   public void run()
   {
      try
      {
         run2();
      }
      catch(IOException e)
      {
         System.err.println("Exception during snoop on client:\n"+target+"\nDetails:");
         System.err.println(e);
         if(open) try{sock.close(); open = false;}catch(IOException e2){}
      }
   }

   public void run2() throws IOException
   {
      sock = new Socket(target.getAddress(), target.getSnoopPort());
      open = true;
      // set socket options
      sock.setTcpNoDelay(CacheConstants.TCPNODELAY);
      sock.setSoTimeout(CacheConstants.TIMEOUT);
      sock.setSoLinger(CacheConstants.LINGER, CacheConstants.LINGERTIME);
      // imposto gli stream
      datain = new DataInputStream( sock.getInputStream() );
      dataout = new DataOutputStream( sock.getOutputStream() );
      // invia l'ID del server ed il comando
      long serverID = owner.getInfo().getID();
      dataout.writeLong(serverID);
      short command = inv_snoop ? CacheConstants.SNOOP_INV :
                                  CacheConstants.SNOOP ;
      dataout.writeShort(command);
      // invia il tag
      tag.writeImage(dataout);
      // flush del buffer
      dataout.flush();
      // attendi risultato
      res[index] = datain.readShort(); 
      switch(res[index])
      {
      case CacheConstants.MISS:
      case CacheConstants.HIT:
         elemres[index] = null; break;
      case CacheConstants.HITM:
         elemres[index] = owner.readCacheElem(datain);
         break;
      }
      
      // invia l'ack
      dataout.writeShort(CacheConstants.ACK);
      dataout.flush();
      
      // chiude la socket
      sock.close();
      open = false;                                 
   }
                                 
} // CacheSnooperThread