/** File CacheStorageServer.java
* Classe astratta Java che implementa  
* un server di una cache dotato di memoria.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;

public abstract class CacheStorageServer extends CacheServer
{
   // membri
   // unit di memoria
   protected IStorageUnit memcard = null;
   
   public CacheStorageServer( String servername, 
                              long serverID,
                              int serverport,
                              boolean masterserver,
                              ILispList brothersInfo,
                              int priority ) throws IOException
   {
      super( servername, serverID, serverport,
             masterserver, brothersInfo, priority );
   }
   
   public void run()
   {
      try
      {
         memcard = buildMemoryUnit();      
      }
      catch(IOException e)
      {
         System.err.println("Unable to load memory unit.");
         System.err.println(e);
         e.printStackTrace();
         return;
      }
      super.run();
   }
   
   protected abstract IStorageUnit buildMemoryUnit() throws IOException;
   
   protected CacheDataInfo LoadData( ITag tag ) throws 
                                                DataIsMissingException,
                                                ServerPermissionException,
                                                IOException
   { return memcard.LoadData(tag); }                                                
      
   // serve a caricare l'ultimo valore valido rimasto sul server
   // a differenza di loaddata la DataIsMissingException 
   // lanciata solo se proprio il dato NON c'
   public CacheDataInfo LoadLastValidValue( ITag tag ) throws 
                                                DataIsMissingException,
                                                ServerPermissionException,
                                                IOException
   { return memcard.LoadLastValidValue(tag); }
                                                   
   public void SaveDataToStorage(CacheDataInfo dat) throws IOException
   {
      memcard.SaveDataToStorage(dat);
   }
   
   protected void MarkDataAsInvalid(ITag tag, CacheServerInfo newsource)
                                                throws IOException
   {
      memcard.MarkDataAsInvalid(tag, newsource);
   }                                                
                                                   
   protected void MarkDataAsInvalidFromCache(ITag tag, CacheClientInfo newsource)
                                                throws IOException
   {
      memcard.MarkDataAsInvalidFromCache(tag, newsource);
   }                                                
                                                
   protected void RestoreMarkedAsInvalidDatas() throws IOException
   {
      memcard.RestoreMarkedAsInvalidDatas();
   }
   
   protected void TerminateMaster()
   {
      super.TerminateMaster();
      try
      {
         System.out.println("Saving all unsaved datas...");
         memcard.closeUnit();
         //debug
         System.err.println("All the datas have been saved.");
      }
      catch(IOException e)
      {
         System.err.println("Error during closing procedure.\nDetails follow...");
         System.err.println(e);
      }
   }

   // metodo di CacheServer
   protected abstract CacheCLMDNR createConsistencyController();
   
} // CacheStorageServer

