/** File HashStorageUnit.java
* Classe Java che implementa  
* la memoria del server di una cache.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;
import java.util.*;

public class HashStorageUnit implements IStorageUnit
{
   private CacheHashStorageServer owner = null;
   private Hashtable htable = null;
   private Hashtable invtable = null;
   
   protected HashStorageUnit( CacheHashStorageServer owner )
                                    throws IOException
   {
      this.owner = owner;
      htable = new Hashtable();
      invtable = new Hashtable();
      // solo il master ha una memoria iniziale
      if(!owner.getMaster()) return;
      CacheDataInfo[] elements = owner.UploadDataFromDisk();
      for( int i = 0; i< elements.length ; i++ )
      {
         KeyTag key = new KeyTag( elements[i].getElem().getTag() );
         htable.put(key, elements[i]);
      }
   }

   public synchronized CacheDataInfo LoadData( ITag tag ) throws 
                                                DataIsMissingException
   {
      CacheDataInfo res = LoadLastValidValue(tag);
      KeyTag key = new KeyTag(tag);
      Object obj = invtable.get(key);// se il dato  invalido
      if(obj != null) throw new DataIsMissingException(tag);
      else return res;
   }                                             
      
   // serve a caricare l'ultimo valore valido rimasto sul server
   // a differenza di loaddata la DataIsMissingException 
   // lanciata solo se proprio il dato NON c'
   public synchronized CacheDataInfo LoadLastValidValue( ITag tag ) throws 
                                                DataIsMissingException
   {
      KeyTag key = new KeyTag(tag);
      CacheDataInfo res = (CacheDataInfo)(htable.get(key));
      if(res == null) throw new DataIsMissingException(tag);
      else return res;
   }
                                                   
   public synchronized void SaveDataToStorage(CacheDataInfo dat)
   {
      KeyTag key = new KeyTag(dat.getElem().getTag());
      htable.remove(key);
      invtable.remove(key);
      htable.put(key, dat);
   }
   
   public synchronized void MarkDataAsInvalid(ITag tag, CacheServerInfo newsource)
   {
      KeyTag key = new KeyTag(tag);
      invtable.remove(key);
      invtable.put(key, newsource);
   }
                                                      
   public synchronized void MarkDataAsInvalidFromCache(ITag tag, CacheClientInfo newsource)
   {
      KeyTag key = new KeyTag(tag);
      invtable.remove(key);
      invtable.put(key, newsource);
   }
                                                
   public synchronized void RestoreMarkedAsInvalidDatas()
   {
      invtable.clear();
   }
   
   public synchronized void closeUnit() throws IOException
   {
      // solo il master ha una memoria iniziale
      if(!owner.getMaster()) return;
      
      owner.BeginSavingSession(this);
      
      Enumeration en = htable.elements();
      while(en.hasMoreElements())
      {
         try
         {
            CacheDataInfo dat = (CacheDataInfo)(en.nextElement());
            owner.SaveToDisk(dat);
         }
         catch(IOException e)
         {
            owner.EndSavingSession(this);
            throw e;
         }
      }
      
      owner.EndSavingSession(this);
   }

   // classi per uso interno
   private class KeyTag implements ITag
   {
      private ITag tag = null;
      private int hash = 0;
      
      public KeyTag(ITag tag)
      {
         this.tag = tag;
         hash = owner.getHashCodeFromTag(tag); 
      }
      
      // metodi di ITag
      public String toString()
      {
         return "{ HashCode = "+ hash +" ; tag = "+ tag.toString()+" }";
      }   
      
      public boolean isEqual( ITag toCheck )
      {
         return tag.isEqual(toCheck);
      }

      // metodi di IRawData
      public int writeImage( DataOutputStream dataout ) { return 0; }
      
      public int hashCode() { return hash; }
      
      public boolean equals(Object obj)
      {
         return ( obj != null )&&( obj instanceof KeyTag )&&
                  ((KeyTag)obj).isEqual(tag);
      } 
   } // KeyTag class

} // HashStorageUnit
