/** File LispUtils.java
* Classe statica Java che permette di usare la lista LISP
* come se fosse un contenitore.
* NOTA: Una lista LISP  SEMPRE UN VALORE !!
@author Claudio Forlivesi
@version 1.0, 6/11/1999 
*/

public class LispUtils
{
   public static ILispList EmptyList() { return new LispEmptyList() ; }
   public static ILispList cons( ILispBase obj, ILispList tailList )
   {
      return new LispList( obj, tailList );
   }
   public static ILispBase head( ILispList list ) { return list.Head() ; }
   public static ILispList tail( ILispList list ) { return list.Tail() ; }
   public static ILispList doListFromBase( ILispBase obj )
   {
      return cons( obj, EmptyList() ) ;
   }
   public static ILispList doListFromArray( ILispBase[] ar )
   {
      return doListFromArrayRic( ar, 0 );
   }
   private static ILispList doListFromArrayRic( ILispBase[] ar, int i )
   {
      return ( i < ar.length )?
      /* i < ar.length */ cons( ar[i], doListFromArrayRic( ar, i+1 ) ) :
      /* i == ar.length */ EmptyList();
   }
   public static ILispList InsertElemAtPos( ILispList list, ILispBase elem, int pos )
   {
      return ( pos == 0 )? cons( elem, list ) :
               cons( head(list), InsertElemAtPos(tail(list), elem, pos-1) ) ;
   }
   public static int getListSize( ILispList list )
   {
      return ( list instanceof LispEmptyList )? 0 : 1+getListSize(tail(list)) ;
   }
   public static ILispList InsertHead( ILispList list, ILispBase elem )
   {
      return cons( elem, list );
   }
   public static ILispList InsertTail( ILispList list, ILispBase elem )
   {
      return ( list instanceof LispEmptyList )? cons( elem, list ) :
                                 cons(head(list), InsertTail(tail(list), elem)) ;
   }
   public static ILispList InsertOrdered( ILispList list, ILispBase elem )
   {
      return (( list instanceof LispEmptyList )||( elem.isLess( head(list) ) )) ? 
      																			 cons( elem, list):
									cons( head(list), InsertOrdered( tail(list), elem ) ) ;
   }
   public static ILispList Reversed( ILispList list )
   {
      return (tail(list) instanceof LispEmptyList)? list :
             InsertTail( Reversed(tail(list)), head(list) ) ;
   }
   public static ILispList OrderLowFirst( ILispList list )
   {
      return ( tail(list) instanceof LispEmptyList )? list :
                 MergeNOrder( EmptyList(), list );
   }
   public static ILispList OrderHighFirst( ILispList list )
   {
      return Reversed(OrderLowFirst(list)) ;
   }
   public static ILispBase getElemAtPos( ILispList list, int pos )
   {
      return (pos == 0)? head(list) : getElemAtPos( tail(list), pos-1 ) ;
   }
   public static ILispBase getLastElem( ILispList list )
   {
      return ( tail(list) instanceof LispEmptyList )? head(list) : getLastElem(tail(list)) ;
   }
   public static int getElemPos( ILispList list, ILispBase elem )
   {
      return ( list instanceof LispEmptyList )? -1 : 
           ( ( head(list).isEqual(elem) )? 0 : 
              ( (getElemPos(tail(list),elem) == -1)? -1 : 1+getElemPos(tail(list),elem) ) ) ;
   }
   public static int getElemPosUnsafe( ILispList list, ILispBase elem ) throws Exception
   {
      if ( list instanceof LispEmptyList ) throw new Exception("ERRORE! Tipo: Exception \nCausa: L'elemento "+elem+"non  stato trovato nella lista :\n"+list) ;
      
      return (head(list).isEqual(elem)) ? 0 : 1+getElemPos(tail(list),elem) ;
   }
   public static ILispList RemoveHead( ILispList list )
   {
      return tail(list);
   }
   public static ILispList RemoveTail( ILispList list )
   {
      return ( tail(list) instanceof LispEmptyList ) ? tail(list) :
               cons( head(list), RemoveTail(tail(list)) ) ;
   }
   public static ILispList RemoveElemAtPos( ILispList list, int pos )
   {
      return ( pos == 0 )? tail( list ) :
               cons( head(list), RemoveElemAtPos(tail(list), pos-1) ) ;
   }
   public static ILispList RemoveElem( ILispList list, ILispBase elem )
   {
      return ( list instanceof LispEmptyList ) ? EmptyList() :
            (
               (head(list).isEqual(elem)) ? RemoveElem(tail(list),elem) :
                                 cons( head(list), RemoveElem(tail(list), elem) )
              ) ;
   }
   public static ILispList MergeLists( ILispList first, ILispList last )
   {
      return ( first instanceof LispEmptyList )? last :
             ( ( tail(first) instanceof LispEmptyList )? cons( head(first), last ) :
                                                         cons( head(first), MergeLists(tail(first), last ))
              ) ;
   }
   public static ILispList MergeNOrder( ILispList dest, ILispList src )
   {
      return ( src instanceof LispEmptyList )? dest :
            MergeNOrder( InsertOrdered( dest, head(src) ), tail(src) ) ;
   }
   
} // LispUtils

 