/** File MagServer.java
* Classe Java che implementa  
* un server di una cache dotato di memoria 
* formata da una hash table per un magazzino.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;
import java.util.*;

public class MagServer extends CacheHashStorageServer
{
   /* membri ereditati 
   // unit di memoria
      IStorageUnit memcard = null;
   */
   // membri privati
   private File elencomateriali = null;
   private File magazzino = null;
   private BufferedWriter writebr = null;
   private BufferedWriter writebr2 = null;
   
   public MagServer( String servername, 
                     long serverID,
                     int serverport,
                     boolean masterserver,
                     ILispList brothersInfo,
                     int priority,
                     File elencomateriali,
                     File magazzino ) throws IOException
   {
      super( servername,
             serverID, serverport,
             masterserver, brothersInfo, priority );
      this.elencomateriali = elencomateriali;
      this.magazzino = magazzino;
   }
   
   // nuovi metodi astratti
   // questo qui carica i dati da un file all'inizio dell'
   // esecuzione del master
   protected CacheDataInfo[] UploadDataFromDisk()
                                    throws IOException
   {
      BufferedReader br = new BufferedReader( new FileReader(magazzino));
      String line;
      String parsedfile = magazzino.getName();
      int nline = 1;
      ILispList datlist = LispUtils.EmptyList();
      while((line = br.readLine() )!=null)
      {
         StringTokenizer tline = new StringTokenizer(line);
         try
         {
            MagTag tag = new MagTag(tline.nextToken().trim());
            int quantity = Integer.parseInt(tline.nextToken().trim());
            int cost = Integer.parseInt(tline.nextToken().trim());
            StoredMat mat = new StoredMat( tag, quantity, cost );
            CacheDataInfo cdi = new CacheDataInfo( mat,
                                                   LispUtils.EmptyList(),
                                                   LispUtils.EmptyList(),
                                                   null);
            datlist = LispUtils.cons(cdi, datlist);
         }
         catch(NoSuchElementException e)
         {
            System.err.println("Error in parsing " +
            parsedfile + " at line " + nline +".");
            System.err.println(e);
            e.printStackTrace();
            throw new IOException("Parse Error.");
         }
         nline++;
      }
      br.close();
      br = new BufferedReader( new FileReader(elencomateriali));
      parsedfile = elencomateriali.getName();
      nline = 1;
      while((line = br.readLine() )!=null)
      {
         StringTokenizer tline = new StringTokenizer(line);
         try
         {
            String name = tline.nextToken().trim();
            MagTag tag = new MagTag(tline.nextToken().trim());
            String descr = "";
            while(tline.hasMoreTokens())
               descr = descr + " " + tline.nextToken() ;
            MatDescr mat = new MatDescr( name, descr, tag );
            CacheDataInfo cdi = new CacheDataInfo( mat,
                                                   LispUtils.EmptyList(),
                                                   LispUtils.EmptyList(),
                                                   null);
            datlist = LispUtils.cons(cdi, datlist);
         }
         catch(NoSuchElementException e)
         {
            System.err.println("Error in parsing " +
            parsedfile + " at line " + nline +".");
            System.err.println(e);
            e.printStackTrace();
            throw new IOException("Parse Error.");
         }
         nline++;
      }
      br.close();
      int ncount = LispUtils.getListSize(datlist);
      CacheDataInfo[] cdis = new CacheDataInfo[ncount];
      while(!datlist.isNull())
      {
         ncount--;
         cdis[ncount] = (CacheDataInfo)(datlist.car());
         datlist = datlist.Tail();
      }
      
      return cdis;
   }

   // questo fornisce l'hashcode dei dati in base 
   // al loro tag
   protected int getHashCodeFromTag(ITag tag)
   {
      if((tag instanceof MagTag)||(tag instanceof DescrTag))
         return 
            tag.toString().charAt(tag.toString().length()-1)-'0' +
            tag.toString().charAt(tag.toString().length()-2)-'0'*10+
            tag.toString().charAt(tag.toString().length()-3)-'0'*100;
      return 0; 
   }   
   
   // questo qui comincia a salvare i dati su disco
   protected void BeginSavingSession(HashStorageUnit source)
                                          throws IOException
   {
      writebr = new BufferedWriter(new FileWriter(magazzino));
      writebr2 = new BufferedWriter(new FileWriter(elencomateriali));
   }
                                    
   // questo qui salva un dato su disco 
   // durante una sessione di salvataggio
   protected void SaveToDisk(CacheDataInfo dat)
                                       throws IOException
   {
      if(dat.getElem() instanceof StoredMat)
      {
         StoredMat mat = (StoredMat)(dat.getElem());
         String toSave = mat.getCode().toString() + " " +
                         mat.getQuantity() + " " +
                         mat.getCost() ;
         writebr.write(toSave, 0, toSave.length());
         writebr.newLine();
      }
      if(dat.getElem() instanceof MatDescr)
      {
         MatDescr des = (MatDescr)(dat.getElem());
         String toSave = des.getName() + " " +
                         des.getCode().toString() + " " +
                         des.getDescr() ;
         writebr2.write(toSave, 0, toSave.length());
         writebr2.newLine();
      }
   }
                                             
   // questo qui termina di salvare i dati su disco
   protected void EndSavingSession(HashStorageUnit source)
                                       throws IOException
   {
      writebr.close();
      writebr = null;
      writebr2.close();
      writebr2 = null;
   }                                                                        
                                    
   // metodo di CacheServer
   protected CacheCLMDNR createConsistencyController()
   {
      return new MagController();
   }
   
} // CacheHashStorageServer


