/** File MagServerApp.java
* Classe Java che implementa l'applicazione server 
* di una cache dotato di memoria 
* formata da una hash table per un magazzino.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class MagServerApp
{
   // membri
   private static Thread tserver = null;
   private static MagServer server = null;
   private static ServerFrame frame = null;
   
   private static class ServerFrame extends JFrame
   {
      public ServerFrame(String servername) 
      {
         super();
         setTitle("Close this window to stop "+servername +
                  " server."); 
         addWindowListener(new WindowAdapter()
         {
            public void windowClosing(WindowEvent e)
            {
               server.StopServerThreads(false);// fine provocata localmente
               try
               {
                  tserver.join();
               }
               catch(InterruptedException er){};
               System.out.println("Server has been closed.");
               System.exit(0);
            }
         });
         Toolkit tk = Toolkit.getDefaultToolkit();
         Dimension d = tk.getScreenSize();
         setSize( d.width/2, d.height/15 );
      }
   }
   
   private static class ServerWndCloser extends MagServer
   {
      public ServerWndCloser( String servername, 
                              long serverID,
                              int serverport,
                              boolean masterserver,
                              ILispList brothersInfo,
                              int priority,
                              File elencomateriali,
                              File magazzino ) throws IOException
      {
         super( servername,
                serverID, serverport,
                masterserver, brothersInfo, priority,
                elencomateriali, magazzino );
      }
      
      public void run()
	   {
	      super.run();
         if(frame.isShowing()&&endcausedfrommasterserver&&!getMaster())
         {
            //debug
		      System.err.println("Disposing window.");
		      frame.hide();
            //debug
            System.err.println("Disposed.");
            System.exit(0);
         }
	   }
   }
   
   public static void main( String[] args )
   {
      if((args == null)||(args.length == 0))
      {
         System.err.println("Sorry, but you must specify the config file.");
         return;
      }
      
      try
      {
         File config = new File(args[0]);
         BufferedReader br = new BufferedReader(
                                 new FileReader(config));
         // nome del server
         String line = readLine(br);
         StringTokenizer st = new StringTokenizer(line);
         String name = st.nextToken();
         // ID del server
         line = readLine(br);
         st = new StringTokenizer(line);
         long ID = Long.parseLong(st.nextToken());
         // porta del server
         line = readLine(br);
         st = new StringTokenizer(line);
         int port = Integer.parseInt(st.nextToken());
         // priorit del server
         line = readLine(br);
         st = new StringTokenizer(line);
         int pri = Integer.parseInt(st.nextToken());
         // master o slave
         line = readLine(br);
         st = new StringTokenizer(line);
         String mst = st.nextToken();
         boolean master;
         if((!mst.trim().toLowerCase().equals("master"))&&
            (!mst.trim().toLowerCase().equals("slave")))
            throw new Exception("Server must be a master or a slave");
         if(mst.trim().toLowerCase().equals("master"))
            master = true;
         else master = false;
         
         File elenco = null;
         File magazz = null;
         
         if(master)
         {
            // elenco dei materiali
            line = readLine(br);
            st = new StringTokenizer(line);
            elenco = new File(st.nextToken());
            // db del magazzino
            line = readLine(br);
            st = new StringTokenizer(line);
            magazz = new File(st.nextToken());
         }
         
         // elenco dei server slaves o del master
         ILispList bros = LispUtils.EmptyList();
         if(master)
         {
            // info sugli slaves
            CacheServerInfo sinfo = null;
            while(null!=(sinfo = getNextServerInfo(br)))
               bros = adjustServerList( sinfo, bros );
         }
         else
         {
            // info sul master
            CacheServerInfo minfo = getNextServerInfo(br);
            if(minfo == null)
               throw new Exception("Unable to locate or find the master server");
            bros = adjustServerList(minfo, bros);   
         }
         
         // fine della config
         br.close();
         // crea il server
         frame = new ServerFrame(name);
         server = new ServerWndCloser( name, ID, port,
                                       master, bros, pri,
                                       elenco, magazz );
         
         tserver = new Thread(server);
         tserver.start();
         frame.show();
      }
      catch(FileNotFoundException fe)
      {
         System.err.println("Unable to create the server.");
         System.err.println("File "+ args[0] +" not found.");
         System.err.println(fe);
         fe.printStackTrace();
      }
      catch(Exception e)
      {
         System.err.println("Unable to create the server.");
         System.err.println(e);
         e.printStackTrace();
      }
      //debug
      System.err.println("main() ends.");
   }
   
   private static final CacheServerInfo getNextServerInfo(BufferedReader br)
   {
      String name = null;
      String log_addr = null;
      try
      {
         // nome del server
         String line = readLine(br);
         StringTokenizer st = new StringTokenizer(line);
         name = st.nextToken();
         // ID del server
         line = readLine(br);
         st = new StringTokenizer(line);
         long ID = Long.parseLong(st.nextToken());
         // porta del server
         line = readLine(br);
         st = new StringTokenizer(line);
         int port = Integer.parseInt(st.nextToken());
         // priorit del server
         line = readLine(br);
         st = new StringTokenizer(line);
         int pri = Integer.parseInt(st.nextToken());
         // indirizzo del server
         line = readLine(br);
         st = new StringTokenizer(line);
         log_addr = st.nextToken();
         InetAddress addr = InetAddress.getByName(log_addr);
         
         return new CacheServerInfo(name, ID, addr, port, pri);
      }
      catch(UnknownHostException ue)
      {
         System.err.println("Unable to locate server " +
                            name +" at " + log_addr +
                            " \nThis server won't be recognized in server operations.");
         return null;
      }
      catch(Exception e)
      {
         if(name == null) return null;
         // se invece  stato letto il nome di un server
         System.err.println(e);
         e.printStackTrace();
         System.err.print("This exception will cause server"+name+
         " not to be recognized during server operations.\n");
         
         return null;
      }
   }
   
   private static String readLine(BufferedReader br) throws IOException
   {
      String ret = br.readLine();
      if(null==ret) throw new IOException("No More lines exception");
      return ret;
   }
   
   private static final ILispList adjustServerList(CacheServerInfo news, 
                                                   ILispList oldlist)
   {
      return LispUtils.InsertTail(oldlist, news);
   }                                                   
   
} // MagServerApp

