/** File MyPanel.java
* Classe astratta Java che implementa 
* le basi dell'interfaccia utente 
* per gestire un magazzino.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.util.*;

public class MyPanel
{
   // costanti
   protected static final int EXIT_OPTION = 7;
   protected static final String[] elementfields =
            {"NAME or CODE", 
             "QUANTITY (in units)",
             "UNIT COST (in $)" };
   protected static final String[] costfields =
            {"Name or Code",
             "Cost new=1,add=2,sub=3", 
             "Amount (in $)" };
   protected static final String[] quantityfields =
            {"Name or Code",
             "Quantity new=1,add=2,sub=3", 
             "Amount (in units)" };
   protected static final String[] desccolumns =
            {"NAME", "CODE", "DESCRIPTION"};
   protected static final String[] options =
   {
      "1)See description of an element",
      "2)See information about a stored voice",
      "3)Insert a new element description",
      "4)Edit quantity/cost of a stored voice",
      "5)Update cost of a voice",
      "6)Update quantity of a voice",
      "7)Exit"
   };
   
   // membri
   protected CacheMag cache = null;
   protected boolean appmustexit = false;
   
   // costruttore
   public MyPanel(CacheMag cache)
   {
      this.cache = cache;
   }
   
   // metodi
   public void display()
   {
      ShowInterface();
      int opt;
      do
      {
         UpdateInterface();
         opt = getOption();
         InterpretateOption(opt);
      }
      while((opt != EXIT_OPTION)&&!appmustexit);
      DisposeInterface();
   }
   
   protected void exitApplication()
   {
      appmustexit = true;
   }
   
   protected void resumeApplication()
   {
      appmustexit = false;
   }
   
   protected void InterpretateOption(int option)
   {
      switch(option)
      {
      case 1: ShowElemDescr(); break;
      case 2: ShowInfoAboutAVoice(); break;
      case 3: NewElementDescr(); break;
      case 4: NewStoredVoice(); break;
      case 5: UpdateCost(); break;
      case 6: UpdateQuantity(); break;
      }
   }
   
   protected void ShowElemDescr()
   {
      String name = UserQuery("Please insert element name:");
      DescrTag tag = new DescrTag(name);
      try
      {
         // leggi l'elemento e mettilo in cache
         ICacheElem answer = cache.read(tag, true);
         MatDescr descr = (MatDescr)answer;
         String title = "Description on material "+descr.getName();
         String[] columns = 
            {"NAME", "CODE", "DESCRIPTION"} ;
         String[][] content =
         {
            {descr.getName(),descr.getCode(),descr.getDescr()}
         };
         TableOutput(title, columns, content);
      }
      catch(Exception ce)
      {
         ExceptionOutput(ce);
         if(ce instanceof ServerDownException)
            exitApplication();
      }
   }
   
   protected void ShowInfoAboutAVoice()
   {
      String namecode = UserQuery("Please insert element name or code:");
      // prova se  un nome
      DescrTag tag = new DescrTag(namecode);
      boolean tryasname = true;
      MatDescr descr = null;
      try
      {
         // leggi l'elemento e mettilo in cache
         descr = (MatDescr)cache.read(tag, true);
      }
      catch(Exception ce)
      {
         tryasname = false;
      }
      
      try
      {
         MagTag mtag;
         if(tryasname)
            mtag = new MagTag(descr.getCode());
         else mtag = new MagTag(namecode);
         // leggi l'elemento e mettilo in cache
         StoredMat mat = (StoredMat)cache.read(mtag, true);
         
         String title = "Description on material "+namecode;
         String[] columns = 
            {"CODE", "QUANTITY", "UNIT COST"};
         String[][] content =
         {
            {mat.getCode().toString(),
             "" + mat.getQuantity() + " units",
             "" + mat.getCost() + "$/unit" }
         };
         TableOutput(title, columns, content);
      }
      catch(Exception ce)
      {
         ExceptionOutput(ce);
         if(ce instanceof ServerDownException)
            exitApplication();
      }
   }
   
   protected void NewElementDescr()
   {
      String[] userans =
         MultipleUserQuery("Please fill the fields below.",
                           desccolumns);
      MatDescr mdesc = new MatDescr( userans[0], // name
                                     userans[2], // desc
                                     userans[1]);// code
      try
      {
         cache.write(mdesc, true);
      }
      catch(Exception ce)
      {
         ExceptionOutput(ce);
         if(ce instanceof ServerDownException)
            exitApplication();
      }
   }
   
   protected void NewStoredVoice()
   {
      String[] userans =
         MultipleUserQuery("Please fill the fields below.",
                           elementfields);
      int quantity=0;
      int cost=0;
      String namecode = userans[0];
      // prova se  un nome
      DescrTag tag = new DescrTag(namecode);
      boolean tryasname = true;
      MatDescr descr = null;
      try
      {
         quantity = Integer.parseInt(userans[1]);
         cost = Integer.parseInt(userans[2]);
         // leggi l'elemento e mettilo in cache
         descr = (MatDescr)cache.read(tag, true);
      }
      catch(NumberFormatException ne)
      {
         MessageOutput("Numeric format incorrect.");
         return;
      }
      catch(Exception ce)
      {
         tryasname = false;
      }
      
      try
      {
         MagTag mtag;
         if(tryasname)
            mtag = new MagTag(descr.getCode());
         else mtag = new MagTag(namecode);
         
         StoredMat mat = new StoredMat( mtag,
                                        quantity,
                                        cost );
         cache.write(mat, true);
      }
      catch(Exception ce)
      {
         ExceptionOutput(ce);
         if(ce instanceof ServerDownException)
            exitApplication();
      }
   }
   
   protected void UpdateCost()
   {
      String[] userans =
         MultipleUserQuery("Please fill the fields below.",
                           costfields);
      short oper=0;
      int opcost=0;
      String namecode = userans[0];
      // prova se  un nome
      DescrTag tag = new DescrTag(namecode);
      boolean tryasname = true;
      MatDescr descr = null;
      try
      {
         oper = Short.parseShort(userans[1]);
         opcost = Integer.parseInt(userans[2]);
         if((oper<1)||(oper>3))
         {
            MessageOutput("Unknown OpCode.");
            return;
         }
         // leggi l'elemento e mettilo in cache
         descr = (MatDescr)cache.read(tag, true);
      }
      catch(NumberFormatException ne)
      {
         MessageOutput("Numeric format incorrect.");
         return;
      }
      catch(Exception ce)
      {
         tryasname = false;
      }
      
      try
      {
         String code;
         if(tryasname)
            code = descr.getCode();
         else code = namecode;
         UpdateOp upd = new UpdateOp(code, 
             (short)(MagConstants.UPDATEOP | oper),
                                     opcost);
         
         cache.update(upd, true);
      }
      catch(Exception ce)
      {
         ExceptionOutput(ce);
         if(ce instanceof ServerDownException)
            exitApplication();
      }
   }
         
   protected void UpdateQuantity()
   {
      String[] userans =
         MultipleUserQuery("Please fill the fields below.",
                           quantityfields);
      short oper=0;
      int opquant=0;
      String namecode = userans[0];
      // prova se  un nome
      DescrTag tag = new DescrTag(namecode);
      boolean tryasname = true;
      MatDescr descr = null;
      try
      {
         oper = Short.parseShort(userans[1]);
         opquant = Integer.parseInt(userans[2]);
         if((oper<1)||(oper>3))
         {
            MessageOutput("Unknown OpCode.");
            return;
         }
         else oper = (short)(oper * 4);
         // leggi l'elemento e mettilo in cache
         descr = (MatDescr)cache.read(tag, true);
      }
      catch(NumberFormatException ne)
      {
         MessageOutput("Numeric format incorrect.");
         return;
      }
      catch(Exception ce)
      {
         tryasname = false;
      }
      
      try
      {
         String code;
         if(tryasname)
            code = descr.getCode();
         else code = namecode;
         UpdateOp upd = new UpdateOp(code, 
             (short)(MagConstants.UPDATEOP | oper),
                                     opquant,0);
         
         cache.update(upd, true);
      }
      catch(Exception ce)
      {
         ExceptionOutput(ce);
         if(ce instanceof ServerDownException)
            exitApplication();
      }
   }
   
   protected void ShowInterface()
   {
      System.out.println("Available options:");
   }
   
   protected void ExceptionOutput(Exception e)
   {
      System.err.println(e);
      e.printStackTrace();
   }
   
   protected void MessageOutput(String msg)
   {
      System.out.println(msg);
   }
   
   protected void TableOutput(String title, 
                              String[] columns,
                              String[][] content)
   {
      System.out.println(title);
      int i;
      int j;
      for( i= 0; i < columns.length ; i++)
      {
         System.out.print(columns[i]);
         if(i != columns.length -1)
            System.out.print("\t");
         else System.out.print("\n");
      }
      for( i= 0; i < content.length ; i++)
         for( j= 0; j < content[i].length ; j++)
         {
            System.out.print(content[i][j]);
            if(j != content[i].length -1)
               System.out.print("\t");
            else System.out.print("\n");
         }
   }
   
   protected String[] MultipleUserQuery(String question, 
                                        String[] fields)
   {
      System.out.println(question);
      String[] ret = new String[fields.length];
      for( int i = 0 ; i < fields.length ; i++ )
      {
         ret[i] = UserQuery(fields[i]+"=");
      }
      return ret;
   }
   
   protected void UpdateInterface()
   {
      for(int i = 0 ; i < options.length ; i++)
         System.out.println(options[i]);
   }

   protected void DisposeInterface()
   {
      System.out.println("Exit option selected.");
   }
   
   protected int getOption()
   {
      String line = null;
      try
      {
         line =
         (new BufferedReader(
            new InputStreamReader(System.in))).readLine();
         String first = (new StringTokenizer(line)).nextToken().trim();
         return Integer.parseInt(first);
      }
      catch(Exception e)
      {
         return EXIT_OPTION;
      }
   }

   protected String UserQuery(String query)
   {
      System.out.print(query);
      try
      {
         BufferedReader br = 
            new BufferedReader(
               new InputStreamReader(
                  System.in));
         String ret = br.readLine();
         if (ret == null ) return "";
         else return ret;
      }
      catch(Exception e)
      { return ""; }
   }

} // MyPanel
