/** File ServerServingThread.java
* Classe Java che implementa  
* un thread del server che interagisce con un altro server.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;

public class ServerServingThread extends ServingThread 
{
   public ServerServingThread( Socket s, CacheServer o,
                               DataInputStream datain,
                               DataOutputStream dataout )
                                       throws IOException
   {
      super(s, o, datain, dataout);
   }
   
   /* dati protected ereditati
   protected DataInputStream datain;
   protected DataOutputStream dataout;
   protected Socket sock;
   protected CacheServer owner;
   */
   
   // dati nostri
   protected CacheServerInfo serverinfo = null;
   protected ITag tag = null;
   protected ICacheElem receivedElem = null;
   private ICacheElem[] elemres = null;
   private short[] res = null;
   private ServerSnooperThread[] threads = null;
   private ILispList targets = LispUtils.EmptyList();
   
   public void run()
   {
      long srcID = 0;
      short command = CacheConstants.DETECT_DATA;// il pi innocuo
      boolean open = true; // indica che la socket  attiva
      try
      {
         // leggi l'ID del server
         srcID = datain.readLong();
         // leggi il comando
         command = datain.readShort();
         // just get info on the server source
         ILispList broslist = owner.getBros();
         boolean found = false;
         CacheServerInfo src = null;
         while(!found && !broslist.isNull())
         {
            src = (CacheServerInfo)(broslist.car());
            found = src.getID() == srcID;
            broslist = broslist.Tail();
         }
         // did we found it??
         if(!found) // nay, get lost, please
         {
            sock.close();
            open = false;
            return;
         }
         // if we are here we did. considering command now...
         serverinfo = src;
         switch(command)
         {
         case CacheConstants.WRITE_DATA: writeData(); break;
         case CacheConstants.INV_DATA: invalidateData(); break;
         case CacheConstants.DETECT_DATA: detectData(); break;
         case CacheConstants.GET_DATA: getData(); break;
         case CacheConstants.SLAVE_IS_CLOSING: slaveIsClosing(); break;
         case CacheConstants.MASTER_IS_CLOSING: masterIsClosing(); break;                
         default: unknownCommand();
         }
         
         // end communications, disconnecting...
         sock.close();
         open = false;
      }
      catch(IOException e)
      {
         System.err.println("Error during snoop operation from server: ");
         System.err.println(((serverinfo == null)?"Unknown":serverinfo.toString()));
         System.err.println("Involved data tag:");
         System.err.println(((tag == null)?"Unknown":tag.toString()));
         System.err.println("Details follow...");
         System.err.println(e);
         if(open) try{sock.close(); open = false;}catch(IOException e2){}
         // errore forse sullo slave
         if(owner.getMaster()) owner.errorInSlave(serverinfo);
         // altrimenti c' un errore sul master
         else owner.errorInMaster(serverinfo);
      }
   }
   
   private void slaveIsClosing() throws IOException
   {
      //debug
      System.err.println("Closure of "+serverinfo.getName()+" detected.");
      // semplicemente rimuove lo slave dalla lista dei server
      owner.RemoveServer(serverinfo);
      // risponde con ack
      dataout.writeShort(CacheConstants.ACK);
      dataout.flush();
   }
   
   private void masterIsClosing() throws IOException
   {
      // lancia la chiusura
      owner.StopServerThreads(true);// cause is another server
      // risponde con ack
      dataout.writeShort(CacheConstants.ACK);
      dataout.flush();
   }

   private void writeData() throws IOException
   {
            // devo per forza essere il master
            // infatti lo slave dal master pu
            // ricevere solo invalidazioni
      CacheDataInfo searched = null;
      CacheDataInfo newdata = null;
      boolean miss = false;
      boolean collision = false;

      receivedElem = readCacheElem(datain);
      tag = receivedElem.getTag();
      
      // ottieni accesso al dato
      try
      {
         searched = owner.AskDataForWriting(tag,true);
         miss = searched == null;
         if(miss) // caso di miss
         {
            // questa  una new entry
            // poich siamo il master
            dataout.writeShort(CacheConstants.CAN_DO);
            dataout.flush();
            // se c' una collisione esterna esci
            if(datain.readShort() == CacheConstants.DO_ACTION)
            {
               newdata = new CacheDataInfo( receivedElem,
                                            LispUtils.EmptyList(),
                                            LispUtils.EmptyList(),
                                            null );
               owner.SaveDataToStorage(newdata);
            }
            owner.ReleaseData(tag);
            return;
         }
         // caso di OK, dato presente
         // invalida i clients
         CacheDataInfo clientdata = new CacheDataInfo(searched.getElem(),
                                                      searched.getClientList(),
                                                      LispUtils.EmptyList(),
                                                      null);
         clientdata = setExclusiveUse(clientdata, null, null, false);                                                      
         // lancia una invalidazione su tutti gli slaves
         targets = getTargetServers(searched.getServerList(),
                                    searched.getSource());
         int size = LispUtils.getListSize(targets);
         res = new short[size];
         elemres = new ICacheElem[size];
         threads = new ServerSnooperThread[size];          
         collision = BeginSnoopOnServers( false, // snoop invalidante
                                          new CacheDataInfo(
                                             new FakeData(tag),
                                             LispUtils.EmptyList(),
                                             LispUtils.EmptyList(),
                                             null) );
         if(collision) // collisione durante gli snoop
         {
            dataout.writeShort(CacheConstants.CANT_DO_BECAUSE_COLLISION);
            dataout.flush();
            EndSnoopOnServers(true);
            newdata = new CacheDataInfo( searched.getElem(),
                                         LispUtils.EmptyList(),
                                         searched.getServerList(),
                                         searched.getSource());
            owner.SaveDataToStorage(newdata);
            owner.ReleaseData(tag);
            return;                             
         }
         // caso buono
         dataout.writeShort(CacheConstants.CAN_DO);
         dataout.flush();
         // leggi risposta
         collision = datain.readShort() == CacheConstants.UNDO_ACTION;
         if(collision) // collisione proveniente dallo snooper
         {
            EndSnoopOnServers(true);
            newdata = new CacheDataInfo( searched.getElem(),
                                         LispUtils.EmptyList(),
                                         searched.getServerList(),
                                         searched.getSource());
            owner.SaveDataToStorage(newdata);
            owner.ReleaseData(tag);
            return;                             
         }
         // tutto ok
         EndSnoopOnServers(false);// termina per bene
         newdata = new CacheDataInfo( receivedElem,
                                      LispUtils.EmptyList(),
                                      LispUtils.cons(serverinfo, LispUtils.EmptyList()),
                                      null) ; // poich siamo il master
         owner.SaveDataToStorage(newdata);
         owner.ReleaseData(tag);
         return;                             
      }
      catch(CantSnoopException cte){} // agiamo solo sui clients dunque questa non pu avvenire
      catch(CollisionException ce) 
      {
         // collisione
         dataout.writeShort(CacheConstants.CANT_DO_BECAUSE_COLLISION);
         dataout.flush();
      } 
      catch(ServerPermissionException spe) // facciamo fallire tutto
      {
         // errore
         dataout.writeShort(CacheConstants.CANT_DO_BECAUSE_COLLISION);
         dataout.flush();
      } 
   }
   
   private void invalidateData() throws IOException
   {
      if(owner.getMaster()) // lo slave aggiorna sempre il master
      {
         receivedElem = readCacheElem(datain);
         tag = receivedElem.getTag();
      }
      // altrimenti se siamo uno slave riceveremo solo il tag
      else tag = readTag(datain);

      CacheDataInfo searched = null;
      CacheDataInfo newdata = null;
      boolean miss = false;
      boolean collision = false;

      // ottieni accesso al dato
      try
      {
         searched = owner.AskDataForWriting(tag,true);
         miss = searched == null;
         if(miss) // caso di miss
         {
            // master o slave non fanno nulla
            dataout.writeShort(CacheConstants.DATA_MISS);
            dataout.flush();
         }
         // caso di OK, dato presente
         // master e slave entrambi devono invalidare
         // i clients
         CacheDataInfo clientdata = new CacheDataInfo(searched.getElem(),
                                                      searched.getClientList(),
                                                      LispUtils.EmptyList(),
                                                      null);
         clientdata = setExclusiveUse(clientdata, null, null, false);                                                      
         searched = new CacheDataInfo(searched.getElem(),
                                      LispUtils.EmptyList(),
                                      searched.getServerList(),
                                      searched.getSource());
         // ora per dobbiamo differenziarci
         // tra master e slave
         if(owner.getMaster())
         {
            // lancia una invalidazione su tutti gli slaves
            targets = getTargetServers(searched.getServerList(),
                                       searched.getSource());
            int size = LispUtils.getListSize(targets);
            res = new short[size];
            elemres = new ICacheElem[size];
            threads = new ServerSnooperThread[size];          
            collision = BeginSnoopOnServers( false, // snoop invalidante
                                             new CacheDataInfo(
                                                new FakeData(tag),
                                                LispUtils.EmptyList(),
                                                LispUtils.EmptyList(),
                                                null) );
            if(collision) // collisione durante gli snoop
            {
               dataout.writeShort(CacheConstants.CANT_DO_BECAUSE_COLLISION);
               dataout.flush();
               EndSnoopOnServers(true);
               newdata = new CacheDataInfo( searched.getElem(),
                                            LispUtils.EmptyList(),
                                            searched.getServerList(),
                                            searched.getSource());
               owner.SaveDataToStorage(newdata);
               owner.ReleaseData(tag);
               return;                             
            }
            // caso buono
            dataout.writeShort(CacheConstants.CAN_DO);
            dataout.flush();
            // leggi risposta
            collision = datain.readShort() == CacheConstants.UNDO_ACTION;
            //debug
            System.err.println("Snooper Server told me that ther is"+
                               (collision?" a ":" no ") + "collision.");
            if(collision) // collisione proveniente dallo snooper
            {
               EndSnoopOnServers(true);
               newdata = new CacheDataInfo( searched.getElem(),
                                            LispUtils.EmptyList(),
                                            searched.getServerList(),
                                            searched.getSource());
               owner.SaveDataToStorage(newdata);
               owner.ReleaseData(tag);
               return;                             
            }
            // tutto ok
            EndSnoopOnServers(false);// termina per bene
            newdata = new CacheDataInfo( receivedElem,
                                         LispUtils.EmptyList(),
                                         LispUtils.EmptyList(),
                                         null) ; // poich siamo il master
         }
         else // slave
         {
            // dopo aver invalidato i clienti
            // lancio l'ok e mi preparo all'inv.
            dataout.writeShort(CacheConstants.CAN_DO);
            dataout.flush();
            // leggo la risposta
            collision = datain.readShort() == CacheConstants.UNDO_ACTION;
            //debug
            System.err.println("Snooper Server told me that ther is"+
                               (collision?" a ":" no ") + "collision.");
            if(collision)
            {
               // searched non contiene da prima 
               // pi nessun cliente
               owner.SaveDataToStorage(searched);
               owner.ReleaseData(tag);
               return;
            }
            newdata = new CacheDataInfo( searched.getElem(),
                                         LispUtils.EmptyList(),
                                         LispUtils.EmptyList(),
                                         serverinfo);// info del master
         }
         // in entrambi i casi
         owner.SaveDataToStorage(newdata);
         owner.MarkDataAsInvalid(newdata.getElem().getTag(), serverinfo);
         owner.ReleaseData(tag);
      }
      catch(CantSnoopException cte){} // agiamo solo sui clients dunque questa non pu avvenire
      catch(CollisionException ce) 
      {
         // collisione
         dataout.writeShort(CacheConstants.CANT_DO_BECAUSE_COLLISION);
         dataout.flush();
      } 
      catch(ServerPermissionException spe) // facciamo fallire tutto
      {
         // errore
         dataout.writeShort(CacheConstants.CANT_DO_BECAUSE_COLLISION);
         dataout.flush();
      } 
   }
   
   private void detectData() throws IOException
   {
      detect_ex(true);
   }
   
   private void getData() throws IOException
   {
      //debug 
      System.err.println("Some external idiot asked a read.");
      CacheDataInfo searched = detect_ex(false);
      if(searched != null) // send the data if a hit has come
      {   
         searched.getElem().writeImage(dataout);
         dataout.flush();
         ReleaseDataCorrectly(searched);
      }
   }
   
   private void unknownCommand() throws IOException
   {
      dataout.writeShort(CacheConstants.DATA_MISS);
      dataout.flush();
   }
   
   
   // very private functions
   private CacheDataInfo detect_ex(boolean immediateRelease) throws IOException
   {
      // but first we need the tag
      tag = readTag(datain);
      CacheDataInfo searched = null;// searched == null -> miss or exception happened
      boolean allowed = true;// allowed == true -> no exception happened
      short answer = CacheConstants.DATA_MISS;
      try
      {
         searched = owner.AskDataForReading(tag,false);
         /*
         // se il server la contiene meglio snoopare i suoi clienti
         if(searched!=null)
         {
            CacheDataInfo temp = shareFromClients(searched, null);
            searched = (temp == null)? searched : temp ;
         }
         */
      }
      catch(CollisionException ce) {} // can't happen
      catch(ServerPermissionException spe) { allowed = false; } // seem to be a miss
      // if everything isok then we're enabled
      // and we must release the data
      if(allowed)
      {
         if(searched==null) owner.ReleaseData(tag);
         else if(immediateRelease)
                 ReleaseDataCorrectly(searched);
         if(searched!=null)
            answer = CacheConstants.DATA_PRESENT;
      }
      dataout.writeShort(answer);
      dataout.flush();
      return searched;
   }
   
   private boolean BeginSnoopOnServers(boolean act,
                                       CacheDataInfo elemInfo) 
   {                                       
      ILispList t = targets;
      for ( int i=0; i< threads.length; i++)
      {
         threads[i] = new ServerSnooperThread( i, res, elemres,
                                               (CacheServerInfo)(t.car()),
                                               elemInfo, owner, act, true);
         threads[i].start();
         t = t.Tail();
      }
      boolean collision = false;
      for ( int i=0; i< threads.length; i++)
      {
         collision = threads[i].waitFirstPhaseEnd() || collision ;
      }      
      
      return collision;
   }
   
   // nota : il server NON scrive mai, invalida soltanto
   private void EndSnoopOnServers(boolean collision)
   {
      for (int i = 0; i < threads.length; i++)
      {
         threads[i].startSecondPhase(collision);
      }
      for (int i = 0; i < threads.length; i++)
      {
         try
         {
            threads[i].join();
         }
         catch(InterruptedException  er){};
      }
   }                                                   

  
   private void ReleaseDataCorrectly(CacheDataInfo data)
                                       throws IOException
   {
      ILispList slist = data.getServerList();
      if(-1==LispUtils.getElemPos(slist, serverinfo)) 
      {
         slist = LispUtils.cons(serverinfo, slist);
         CacheDataInfo newdata = new CacheDataInfo(data.getElem(),
                                                   data.getClientList(),
                                                   slist,
                                                   data.getSource());
         // salvo il dato
         owner.SaveDataToStorage(newdata);
      }
      
      // rilascio il dato
      owner.ReleaseData(data.getElem().getTag());
   }
   
   private ILispList getTargetServers(ILispList copie, CacheServerInfo oldsrc)
   {
      if(copie == null) copie = LispUtils.EmptyList();
      ILispList totalServers = (oldsrc==null)? copie :
                                               LispUtils.cons(oldsrc, copie);
      
      // rimuovo eventualmente m stesso
      try{
      totalServers = LispUtils.RemoveElem(totalServers, owner.getInfo());
      } catch(UnknownHostException uk){};
      // rimuovo lo snooper
      totalServers = LispUtils.RemoveElem(totalServers, serverinfo);
      
      return totalServers;
   }     
                                     
} // ServerServingThread
