/** File ServerSnooperThread.java
* Classe Java che implementa  
* un thread che snoopa gli atri server.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;

public class ServerSnooperThread extends Thread
{
   protected DataInputStream datain;
   protected DataOutputStream dataout;
   protected Socket sock;
   protected CacheServer owner;
   protected boolean master;
   protected boolean act;
   protected boolean snoop;
   protected ICacheElem elemToSnoop;
   protected ITag tag;
   protected CacheServerInfo target;
   protected ICacheElem[] newelems;
   protected short[] answers;
   protected int index;
   
   private Event firstend;
   private Event secondstart;
   private boolean collision;
   private boolean operationdone;
   
   public ServerSnooperThread( int index,
                               short[] answers,
                               ICacheElem[] newelems,
                               CacheServerInfo target,
                               CacheDataInfo elemInfo,
                               CacheServer owner,
                               boolean act,
                               boolean snoop )
   {
      this.index = index;
      this.answers = answers;
      this.newelems = newelems;
      this.target = target;
      elemToSnoop = elemInfo.getElem();
      tag = elemToSnoop.getTag();
      this.owner = owner;
      this.master = owner.getMaster();
      this.act = (!snoop&&act)||(snoop&&act&&!master);
      this.snoop = snoop;
      firstend = new Event();
      secondstart = new Event(); 
      collision = false;
   }
   
   public boolean waitFirstPhaseEnd()
   {
      firstend.waitEvent();
      return collision;
   }
   
   public void startSecondPhase( boolean newcollision )
   {
      collision = newcollision;
      secondstart.Happen();
   }
   
   public void startSecondPhase( boolean newcollision, ICacheElem newelem )
   {
   	elemToSnoop = newelem;
	   startSecondPhase(newcollision);
   }
   
   public void run()
   {
      //debug
      System.err.println("Server Snooper Thread starting on "+
                         target + " for data "+tag);
      InetAddress targetaddr = target.getAddress();
      int targetport = target.getPort();
      boolean open = false; 
      operationdone = false;
      try
      {
         sock = new Socket(targetaddr, targetport);
         open = true;
         // settaggio opzioni socket
         sock.setTcpNoDelay(CacheConstants.TCPNODELAY);
         sock.setSoTimeout(CacheConstants.TIMEOUT);
         sock.setSoLinger(CacheConstants.LINGER, CacheConstants.LINGERTIME);
         datain = new DataInputStream( sock.getInputStream() );
         dataout = new DataOutputStream( sock.getOutputStream() );
         // ora invio l'ordine
         dataout.writeShort(CacheConstants.SERVER_BEGINS_SESSION);
         dataout.writeLong(owner.getInfo().getID());
         short command = (act && snoop) ? CacheConstants.WRITE_DATA :
                         (!act && snoop) ? CacheConstants.INV_DATA :
                         (act && !snoop) ? CacheConstants.GET_DATA :
                       /*(!act && !snoop)*/  CacheConstants.DETECT_DATA ;
         // check del master
         if((command == CacheConstants.WRITE_DATA) && master) 
            command = CacheConstants.INV_DATA;// il master non scrive mai per snoop
         
         dataout.writeShort(command);
         // ci differenziamo a second o meno dello snoop
         if(snoop) TwoPhaseSession();
         else ReadingSession();
         sock.close();
         open = false;
      }
      catch(IOException e)
      {
         System.err.println("Exception during snoop on server:\n"+target+"\nDetails:");
         System.err.println(e);
         // se l'azione non era terminata occorrono provvedimenti
         if(!operationdone)
         {
            if(snoop)
            {
               // write or invalidate
               collision = true;
               firstend.Happen();
               newelems[index] = null;
               answers[index] = CacheConstants.MISS;
            }
            else
            {
               // read or detect
               newelems[index] = null;
               answers[index] = CacheConstants.MISS;
            }
         }
         if(open) try{sock.close(); open = false;}catch(IOException e2){}
      }
      //debug
      System.err.println("Server Snooper Thread finished on "+
                         target + " for data "+tag);
   }
   
   private void ReadingSession() throws IOException
   {
      // invio il tag 
      tag.writeImage(dataout);
      dataout.flush();
      short result = datain.readShort();
      newelems[index] = null;
      if(result == CacheConstants.DATA_MISS)
      {
         //debug
         System.err.println("Snoop on "+tag+" results in a miss.");
         answers[index] = CacheConstants.MISS;
      }
      else
      {
         answers[index] = CacheConstants.HIT;
         //debug
         System.err.println("Snoop on "+tag+" results in a hit.");

         if(act)
         {
            //debug
            System.err.println("Snoop on "+tag+" is now reading the stuff.");
            newelems[index] = owner.readCacheElem(datain);
         }
      }
      operationdone = true;
   }
   
   private void TwoPhaseSession() throws IOException
   {
      if(act)//se siamo uno slave
         elemToSnoop.writeImage(dataout);
      else // invio il tag
         tag.writeImage(dataout);
      dataout.flush();
      short result = datain.readShort();
      newelems[index] = null;
      answers[index] = result;
      switch(result)
      {
      case CacheConstants.DATA_MISS:
         collision = false;
         firstend.Happen();
         break;
      case CacheConstants.CAN_DO:
         collision = false;
         firstend.Happen();
         //debug
         System.err.println("Server Snooper Thread :First phase end. Waiting for second one");
         secondstart.waitEvent();
         //debug
         System.err.println("Server Snooper Thread :Second phase begins.");
         // ora collision contiene lo stato esterno
         short command = collision? CacheConstants.UNDO_ACTION:
                                    CacheConstants.DO_ACTION ;
         dataout.writeShort(command);
         dataout.flush();
         // se non ci sono collisioni uno slave aggiorna sempre il master
         // anche per semplici invalidazioni
         break;
      case CacheConstants.CANT_DO_BECAUSE_COLLISION:
      default:
         collision = true;
         firstend.Happen();
      }
      operationdone = true;
   }

} // ServerSnooperThread