/** File ServingThread.java
* Classe astratta Java che implementa  
* la main architecture di un thread del server di una cache.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;

public abstract class ServingThread implements Runnable,
                                               ICacheDataLoader
{
   protected DataInputStream datain;
   protected DataOutputStream dataout;
   protected Socket sock;
   protected CacheServer owner;
   
   public ServingThread( Socket s, CacheServer o,
                         DataInputStream datain,
                         DataOutputStream dataout )
                              throws IOException
   {
      sock = s;
      owner = o;
      this.datain = datain;
      this.dataout = dataout;
   }
   
   public abstract void run();
   
   public CacheDataInfo setExclusiveUse( CacheDataInfo data, 
                                         CacheServerInfo srvsrc,
                                         CacheClientInfo cltsrc,
                                         boolean updateservers ) throws
                                                CantSnoopException                                                 
   {
      return setExclusiveUse(data, srvsrc, cltsrc, updateservers,
                             CacheConstants.SNOOP_TRIES_NO);
   }
                                                   
   public CacheDataInfo setExclusiveUse( CacheDataInfo data, 
                                         CacheServerInfo srvsrc,
                                         CacheClientInfo cltsrc,
                                         boolean updateservers,
                                         int numtries ) throws
                                                CantSnoopException                                                 
   {
      boolean collision = false;
      int lefts = numtries;
      ILispList serverlist = data.getServerList();
      ILispList clientlist = data.getClientList();
      CacheServerInfo info = data.getSource();
      if(info!=null) serverlist = LispUtils.cons(info, serverlist);
      // elimino le sorgenti
      if(srvsrc!=null) serverlist = LispUtils.RemoveElem( serverlist, srvsrc );
      if(cltsrc!=null) clientlist = LispUtils.RemoveElem( clientlist, cltsrc );
      ILispList leftservers = serverlist;
      while( (lefts>0) ) //&& 
             //((!leftservers.isNull())||(!clientlist.isNull())) )
      {
      //debug
      System.err.println("setExclusive use of "+
                         data.getElem().getTag() +
                         "\nLeft tries "+lefts+
                         "\nLeft clients "+clientlist+
                         "\nLeft Servers "+leftservers);
      lefts--;
      int srsize = LispUtils.getListSize(leftservers);
      int clsize = LispUtils.getListSize(clientlist);
      short[] cres = null;
      short[] sres = null;
      ICacheElem[] cltelems = null;
      ICacheElem[] srvelems = null;
      Thread[] srvth = null;
      Thread[] clth = null;
      
      //debug
      System.err.println("setExclusive use of "+
                         data.getElem().getTag() +
                         " creating server snoopers.");
      if(srsize>0)
      {
         sres = new short[srsize];
         srvelems = new ICacheElem[srsize];
         srvth = new Thread[srsize];
         while(srsize>0)
         {
            Thread snoopy = 
                  new ServerSnooperThread(srsize-1,//indice array
                                          sres,
                                          srvelems,
                                          (CacheServerInfo)(serverlist.car()),
                                          data,// elemento da eliminare
                                          owner,
                                          updateservers,// capisce se occorre anche copiare il dato
                                          true);// indica che deve invalidare
            srvth[srsize-1] = snoopy;
            snoopy.start();
            srsize--;
            serverlist = serverlist.Tail();
         }
      }
      //debug
      System.err.println("setExclusive use of "+
                         data.getElem().getTag() +
                         " creating client snoopers.");
      
      if(clsize>0)
      {
         cres = new short[clsize];
         cltelems = new ICacheElem[clsize];
         clth = new Thread[clsize];
         while(clsize>0)
         {
            Thread snoopy = 
                  new CacheSnooperThread(clsize-1,//indice array
                                         cres,// array che indicher HIT,HITM o MISS
                                         cltelems,// array che indicher il risultato
                                         (CacheClientInfo)(clientlist.car()),// info sul cliente
                                         data.getElem(),// elemento da eliminare
                                         owner, // server
                                         true);// indica che deve invalidare
            clth[clsize-1] = snoopy;
            // NON FAR PERTIRE IL THREAD
            clsize--;
            clientlist = clientlist.Tail();
         }
      }
      
      // attende la fine della prima fase di invalidazione
      serverlist = leftservers;
      clientlist = data.getClientList();
      if(cltsrc!=null) clientlist = LispUtils.RemoveElem( clientlist, cltsrc );
      srsize = LispUtils.getListSize(serverlist);
      clsize = LispUtils.getListSize(clientlist);
      //debug
      System.err.println("setExclusive use of "+
                         data.getElem().getTag() +
                         " waiting for first phase end.");
      for(int i = 0 ; i < srsize ; i++)
         collision=((ServerSnooperThread)srvth[i]).waitFirstPhaseEnd() || collision;
      
      //debug
      System.err.println("setExclusive use of "+
                         data.getElem().getTag() +
                         " has detected that collision is "+collision);
      // rimuovi dalla lista dei server quelli missed
      // per il prossimo tentativo
      if(collision)
         leftservers = NewServerList(leftservers, sres);
      // fai ripartire tutti passando lo stato della collisione
      // partono solo i server rimasti
      //debug
      System.err.println("setExclusive use of "+
                         data.getElem().getTag() +
                         " starting second phase.");
      for(int i = 0 ; i < srsize ; i++)
         if(sres[i]!= CacheConstants.MISS)
            ((ServerSnooperThread)srvth[i]).startSecondPhase(collision);
      // attendi
      //debug
      System.err.println("setExclusive use of "+
                         data.getElem().getTag() +
                         " waiting for snoopers to finish.");
      for(int i = 0 ; i < srsize ; i++)
         try
         {
            srvth[i].join();
         }
         catch(InterruptedException ioe){}
         
      if(!collision)
      {
         // i server hanno tutti dato la loro risposta
         // ora tocca ai clients
         //debug
         System.err.println("setExclusive use of "+
                            data.getElem().getTag() +
                            " interacting with clients.");
         for(int i = 0 ; i < clsize ; i++)
            clth[i].start();
         for(int i = 0 ; i < clsize ; i++)
            try
            {
               clth[i].join();
            }
            catch(InterruptedException rert){};

         /* QUESTO CODICE E' PERICOLOSO : L'HO TOLTO PERCHE' A CHI FA
            UNO SNOOP NON INTERESSA UN NUOVO VALORE PER IL DATO
         if(!updateservers){ // se il dato pi recente NON si trova
                             // sul server locale
         // dobbiamo trovare una versione aggiornata del dato
         // se essa esiste
         for(int i = 0 ; i < srsize ; i++)
            if(CacheConstants.MISS != sres[i])
               return new CacheDataInfo(srvelems[i],
                                        LispUtils.EmptyList(),
                                        LispUtils.EmptyList(),
                                        null);
         for(int i = 0 ; i < clsize ; i++)
            if(CacheConstants.MISS != cres[i])
               return new CacheDataInfo(cltelems[i],
                                        LispUtils.EmptyList(),
                                        LispUtils.EmptyList(),
                                        null);
         }
         */
         //debug
         System.err.println("setExclusive use of "+
                            data.getElem().getTag() +
                            " creating data to return.");
         CacheServerInfo s;
         ILispList ssl;
         // caso default : write sul server con aggiornamento
         // accade che il server che lancia lo snoop se  il
         // master MANTIENE la vecchia lista delle copie del dato
         // a cui per aggiunge la vecchia sorgente, ed inserisce 
         // null nella nuova sorgente, visto che  lui
         // invece lo slave MANTIENE TUTTO COME PRIMA, se  lui
         // ad aggiornare il master
         CacheServerInfo ownerInfo = null;
         try
         { ownerInfo = owner.getInfo(); } catch(UnknownHostException trey){};
         if(updateservers)//caso default
         {
            if(owner.getMaster())
            {
               s = data.getSource();
               ssl = (s==null)? LispUtils.RemoveElem(data.getServerList(),ownerInfo) :
                                LispUtils.RemoveElem(LispUtils.cons(s,data.getServerList()),ownerInfo);
               s = null;                                
            }
            else
            {
               s = data.getSource();
               ssl = data.getServerList();
            }
         }
         // caso write-exclusive: si ha che updateserver = false
         // per cui il master invalider gli slaves:
         // alla fine non avr pi alcun slave legato al dato
         // e la sorgente sar lui, cio null.
         // nel caso in cui lo slave invalida il master
         // invece lo slave DEVE mantenere la vecchia sorgente
         // ma pu fregarsene delle vecchie copie
         else // caso write exclusive
         {
            if(owner.getMaster())
            {
               s = null;
               ssl = LispUtils.EmptyList();
            }
            else
            {
               s = data.getSource();
               ssl = LispUtils.EmptyList();
            }
         }
         return new CacheDataInfo(data.getElem(),
                                  LispUtils.EmptyList(),
                                  ssl,
                                  s);
      }
      
      // collisione : attendi un tempo random
      try
      {wait((long)( Math.random() * Math.pow(2.0,CacheConstants.SNOOP_TRIES_NO-lefts) * 
                    CacheConstants.MAX_WAIT_TIME_AFTER_COLLISION ));}
      catch(InterruptedException e){};
      
      } // fine while(lefts>0)
      
      // fine dei tentativi 
      throw new CantSnoopException();
      
   } // setExclusiveUse()
   
   // ritorna null se tutti gli altri clienti danno MISS
   public CacheDataInfo shareFromClients( CacheDataInfo previous,// serve per ottenere i clienti collegati 
                                          CacheClientInfo cltsrc ) 
   {
      short[] res = null;
      ICacheElem[] elemres = null;
      ILispList clients = previous.getClientList();
      // non snoopare dal cliente di snoop
      if( cltsrc != null )
         clients = LispUtils.RemoveElem(clients, cltsrc );
      int numcl = LispUtils.getListSize(clients);
      // se non ci sono altri clienti ritorna null
      if(numcl == 0 ) return null;
      
      // dimensiona arrays
      res = new short[numcl];
      elemres = new ICacheElem[numcl];
      CacheSnooperThread[] threads = new CacheSnooperThread[numcl];
      
      // lancia threads
      ILispList clist = clients;
      for( int i = 0; i < numcl ; i++ )
      {
         CacheSnooperThread snoopy = new CacheSnooperThread(i,
                                                            res,
                                                            elemres,
                                                            (CacheClientInfo)(clist.car()),
                                                            previous.getElem(), // serve per il tag per cercare l'elemento
                                                            owner,
                                                            false); // non invalidare!!
         threads[i] = snoopy;
         snoopy.start();
         clist = clist.Tail();
      }
      // attendi la loro fine
      for(int i = 0 ; i < numcl ; i++)
         try
         {
            threads[i].join();
         }
         catch(InterruptedException rety){}
      
      // ritorna eventuale risposta
      ILispList newclients = LispUtils.EmptyList(); // lista nuova dei clienti del dato
      // inserisci in lista il cliente dello share
      newclients = LispUtils.cons( cltsrc, newclients );
      // aggiungi i clienti che han risposto
      ICacheElem correctelem = null;// risposta pi corretta finora
      // ci si ferma in caso di HITM
      boolean hitmfound = false;
      clist = clients;
      for(int i = 0 ; i < numcl ; i++)
      {
         if(res[i]!=CacheConstants.MISS)
         {
            // aggiungi il cliente
            newclients = LispUtils.cons( clist.car(), newclients );
            // tieni il nuovo dato
            if(!hitmfound)
            {
               hitmfound = res[i] == CacheConstants.HITM ;
               correctelem = elemres[i];
            }
         }
         clist = clist.Tail();
      }
      if(correctelem!=null) // se c' almeno qualcuno con cui condividere
         return new CacheDataInfo( correctelem,
                                   newclients,
                                   previous.getServerList(),
                                   previous.getSource() );
      // se non ce n' allora nulla
      else return null;
   }                                                                                                

   // rimuove dall' elenco i server che non hanno il dato
   private ILispList NewServerList( ILispList oldlist, short[] res )
   {
      ILispList temp = LispUtils.EmptyList();
      for( int i = 0 ; i < res.length ; i++)
      {
         int listindex = res.length -1 - i;
         if ( res[i] != CacheConstants.MISS )
            temp = LispUtils.cons( LispUtils.getElemAtPos(oldlist, listindex), temp );
      }
      return temp;
   }
   
   // ritorna null se tutti gli altri server danno MISS
   // retrievedata = false evita di ricevere anche il dato (quando esso non serve)
   // ed al chiamante arriva un FakeData
   public CacheDataInfo askDataFromServers( ITag tag, boolean retrievedata ) 
   {
      return askDataFromServers( tag, owner.getBros(), retrievedata );
   }
   
   // ritorna null se tutti gli altri server danno MISS
   public CacheDataInfo askDataFromServers( ITag tag, ILispList serverlist, boolean retrievedata) 
   {
      // dato il tag devo chiedere agli altri server il dato 
      // allora far una query a tutti i server elencati nella lista
      // se essa  nulla o vuota guarder tutti i server fratelli
      // faccio cos : per ogni server attivo una connessione
      // lancio il comando server snoop e poi richiedo il dato
      // senza invalidarlo
      if( serverlist == null || serverlist.isNull() )
         serverlist = owner.getBros();
      
      boolean found = false;
      ILispList slist = serverlist;
      while(!found&& !slist.isNull() )
      {
         short[] res = new short[1];
         ICacheElem[] elemres = new ICacheElem[1];
         elemres[0] = null;
         CacheDataInfo dataToask = new CacheDataInfo(new FakeData(tag),
                                                     LispUtils.EmptyList(),
                                                     LispUtils.EmptyList(),
                                                     null);
         Thread th = new ServerSnooperThread(0, res, elemres,
                                             (CacheServerInfo)slist.car(),
                                             dataToask,// visto che non si deve inviare alcundato ma solo ricevere van bene anche i fakedata
                                             owner,// effettivamente adesso non ci serve sapere se siamo il master
                                             retrievedata, // indica se ricevere anche i dati o solo una indicazione della loro provenienza
                                             false); // NON INVALIDARE E NON AGGIORNARE
         th.start();
         try
         {
            th.join();
         }
         catch(InterruptedException rety) {};
         
         //debug
         System.err.println("Snoop has found "+
                            ((null==elemres[0])?"null":elemres[0].toString()));
         
         found = res[0] != CacheConstants.MISS;
         if(found) 
         {
            // in caso di retrieving del dato costruisci il nuovo
            // elemento di cache correttamente. Altrimenti basta conoscere la source del dato
            ICacheElem element = retrievedata? elemres[0]: new FakeData(tag);
            return new CacheDataInfo(element,
                                     LispUtils.EmptyList(),
                                     LispUtils.EmptyList(),// nessun altro server ci ha ancora CHIESTO il dato
                                     (CacheServerInfo)slist.car());// sorgente del dato
         }
         // se non trova nulla vai avanti al prossimo server
         slist = slist.Tail();                                   
      }
      // se siamo qui non abbiamo trovato NULLA !!!!
      return null;
   }
   
   // ritorna null se tutti gli altri server danno MISS
   public CacheDataInfo askDataFromLinkedServers( ITag tag, boolean retrievedata) 
   {
      if(owner.getMaster())
      {
         // il master si comporta diversamente a seconda 
         // del fatto che il dato sia invalido
         // o si tratti di una new entry
         ILispList linkeds = null;
         try
         {
            CacheDataInfo dat = owner.LoadLastValidValue(tag);
            if(dat.getSource()==null)
               linkeds = dat.getServerList();
            else linkeds = LispUtils.cons(dat.getSource(),
                                          dat.getServerList());
         }
         catch(Exception e){};
         if((linkeds == null) || linkeds.isNull())
            // se non abbiamo ancora
            // trovato nulla  una new entry
            return null;
         else
            return askDataFromServers(tag, linkeds, retrievedata);
      }
      else
      {
         // lo slave chiede sempre a tutti
         // in realt pu chiedere solo al master
         return askDataFromServers(tag, retrievedata);
      }
   }

   // metodi di ICacheDataLoader
   public ITag readTag( DataInputStream datain ) throws IOException
   {
      return owner.readTag(datain);
   }
   
   public IUpdateElem readUpdateElem( DataInputStream datain ) throws IOException
   {
      return owner.readUpdateElem(datain);
   }
   
   public ICacheElem readCacheElem( DataInputStream datain ) throws IOException
   {
      return owner.readCacheElem(datain);
   }
   
   public ICacheElem updateIsCorrect( ICacheElem oldelem, IUpdateElem news ) 
                                          throws InconsistentUpdateException
   {
      return owner.updateIsCorrect(oldelem, news);
   }                                          
   
} // ServingThread
